import numpy as np
from numpy import einsum
from Florence.Tensor import trace, Voigt
from .MaterialBase import Material

class IsotropicElectroMechanics_3(Material):
    """Isotropic electromechanical model in terms of Helmoltz energy
        with one nonlinear electrostatic invariant

            W(C,E) = W_n(C) - eps_1/2*J*C**(-1):(E 0 E) + eps_2/2*(E*E)**2 
            W_n(C) = mu/2*(C:I-3) - mu*lnJ + lamb/2*(lnJ)**2
    
        where 0 stands for dyadic/outer product

    """
    
    def __init__(self, ndim, **kwargs):
        mtype = type(self).__name__
        super(IsotropicElectroMechanics_3, self).__init__(mtype, ndim, **kwargs)
        # REQUIRES SEPARATELY
        self.nvar = self.ndim+1
        self.energy_type = "enthalpy"
        self.nature = "nonlinear"
        self.fields = "electro_mechanics"  

        if self.ndim == 2:
            self.H_VoigtSize = 5
        elif self.ndim == 3:
            self.H_VoigtSize = 9

        # LOW LEVEL DISPATCHER
        self.has_low_level_dispatcher = True
        # self.has_low_level_dispatcher = False

    def KineticMeasures(self,F,ElectricFieldx, elem=0):
        from Florence.MaterialLibrary.LLDispatch._IsotropicElectroMechanics_3_ import KineticMeasures
        return KineticMeasures(self, np.ascontiguousarray(F), ElectricFieldx)

    def Hessian(self,StrainTensors,ElectricFieldx=0,elem=0,gcounter=0):

        mu = self.mu
        lamb = self.lamb
        eps_1 = self.eps_1
        eps_2 = self.eps_2

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]


        E = 1.0*ElectricFieldx.reshape(self.ndim,1)
        Ex = E.reshape(E.shape[0])
        EE = np.dot(E,E.T)
        be = np.dot(b,ElectricFieldx).reshape(self.ndim)

        C_Voigt = lamb/J*einsum('ij,kl',I,I) - (lamb*np.log(J) - mu)/J*( einsum('ik,jl',I,I) + einsum('il,jk',I,I) ) + \
            eps_1*( einsum('ij,kl',I,EE) + einsum('ij,kl',EE,I) - einsum('ik,jl',EE,I) - einsum('il,jk',EE,I) - \
                einsum('ik,jl',I,EE) - einsum('il,jk',I,EE) ) + \
            eps_1*(np.dot(E.T,E)[0,0])*0.5*( einsum('ik,jl',I,I) + einsum('il,jk',I,I) - einsum('ij,kl',I,I) )

        C_Voigt = Voigt(C_Voigt,1)

        P_Voigt = eps_1*( einsum('ik,j',I,Ex) + einsum('jk,i',I,Ex) - einsum('ij,k',I,Ex)) 
        
        P_Voigt = Voigt(P_Voigt,1)
            
        E_Voigt = -eps_1*I + 2.*eps_2/J*(2*np.dot(be,be.T)+np.dot(be.T,be)*I)

        # Build the Hessian
        factor = -1.
        H1 = np.concatenate((C_Voigt,factor*P_Voigt),axis=1)
        H2 = np.concatenate((factor*P_Voigt.T,E_Voigt),axis=1)
        H_Voigt = np.concatenate((H1,H2),axis=0)

        return H_Voigt

    def CauchyStress(self,StrainTensors,ElectricFieldx,elem=0,gcounter=0):

        mu = self.mu
        lamb = self.lamb
        eps_1 = self.eps_1

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]
        E = ElectricFieldx.reshape(self.ndim,1)

        stress = 1.0*mu/J*(b-I) + lamb/J*np.log(J)*I + \
            eps_1*(np.dot(E,E.T) - 0.5*np.dot(E.T,E)[0,0]*I) 

        return stress


    def ElectricDisplacementx(self,StrainTensors,ElectricFieldx,elem=0,gcounter=0):
        
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]
        E = ElectricFieldx.reshape(self.ndim,1)

        varepsilon_1 = self.eps_1
        eps_2 = self.eps_2      
        
        be = np.dot(b,E)
        ebe = np.dot(E.T,be)[0,0]

        D = varepsilon_1*E - 2.*eps_2/J*(ebe)*be
        return D


import numpy as np 
from JacobiPolynomials import *
import math

# 1D - LINE
#------------------------------------------------------------------------------------------------------------------#
#------------------------------------------------------------------------------------------------------------------#
#------------------------------------------------------------------------------------------------------------------#
def NormalisedJacobi1D(C,x):
    p = np.zeros(C+2)

    for i in range(0,C+2):
        p[i] = JacobiPolynomials(i,x,0,0)[-1]*np.sqrt((2.*i+1.)/2.)

    return p  


# 2D - TRI
#------------------------------------------------------------------------------------------------------------------#
#------------------------------------------------------------------------------------------------------------------#
#------------------------------------------------------------------------------------------------------------------#

def NormalisedJacobi2D(C,x):

    # Computes the ortogonal base of 2D polynomials of degree less 
    # or equal to C+1 at the point x=(r,s) in [-1,1]^2 (i.e. on the reference quad)


    N = int( (C+2.)*(C+3.)/2. ) 
    p = np.zeros(N)

    r = x[0]; s = x[1]

    # Ordering: 1st increasing the degree and 2nd lexicogafic order
    ncount = 0 # counter for the polynomials order
    # Loop on degree
    for nDeg in range(0,C+2):
      # Loop by increasing i
        for i in range(0,nDeg+1):
            if i==0:
                p_i = 1.;  q_i = 1.  
            else:
                p_i = JacobiPolynomials(i,r,0.,0.)[-1];     q_i = q_i*(1.-s)/2.
            # Value for j
            j = nDeg-i
            if j==0:
               p_j = 1.
            else:
               p_j = JacobiPolynomials(j,s,2.*i+1.,0.)[-1]
            
            # factor = np.sqrt( (2.*i+1.)*(i+j+1.)/2. )
            factor = math.sqrt( (2.*i+1.)*(i+j+1.)/2. )
            p[ncount] = ( p_i*q_i*p_j )*factor

            ncount += 1
    return p 


def NormalisedJacobiTri(C,x):   
    # Computes the ortogonal base of 2D polynomials of degree less 
    # or equal to n at the point x=(xi,eta) in the reference triangle

    xi = x[0]; eta = x[1] 
    if eta==1: 
        r = -1.; s=1.;
    else:
        r = 2.*(1+xi)/(1.-eta)-1.
        s = eta

    return NormalisedJacobi2D(C,np.array([r,s])) 



def GradNormalisedJacobiTri(C,x,EvalOpt=0):

    # Computes the ortogonal base of 2D polynomials of degree less 
    # or equal to n at the point x=(r,s) in [-1,1]^2
    
    N = int((C+2.)*(C+3.)/2.) 
    p = np.zeros(N);
    dp_dxi  = np.zeros(N)
    dp_deta = np.zeros(N)

    r = x[0]; s = x[1] 

    # THIS MAY RUIN THE CONVERGENCE, BUT FOR POST PROCESSING ITS FINE
    if EvalOpt==1:
        if s==1:
            s=0.99999999999999

    xi = (1.+r)*(1.-s)/2.-1
    eta = s

    dr_dxi  = 2./(1.-eta)
    dr_deta = 2.*(1.+xi)/(1.-eta)**2
    # Derivative of s is not needed because s=eta

    # Ordering: 1st increasing the degree and 2nd lexicogafic order
    ncount = 0
    # Loop on degree
    for nDeg in range(0,C+2):
      # Loop increasing i
        for i in range(0,nDeg+1):
            if i==0:
                p_i = 1;  q_i = 1;  dp_i = 0; dq_i = 0
            else:
                p_i = JacobiPolynomials(i,r,0.,0.)[-1]; dp_i = JacobiPolynomials(i-1,r,1.,1.)[-1]*(i+1.)/2.    
                q_i = q_i*(1.-s)/2.; dq_i = 1.*q_i*(-i)/(1-s)
            
            # Value for j
            j = nDeg-i
            if j==0:
                p_j = 1;  dp_j = 0
            else:
                p_j = JacobiPolynomials(j,s,2.*i+1.,0.)[-1]; dp_j = JacobiPolynomials(j-1,s,2.*i+2.,1.)[-1]*(j+2.*i+2.)/2.  
            
            # factor = np.sqrt( (2.*i+1.)*(i+j+1.)/2. )
            factor = math.sqrt( (2.*i+1.)*(i+j+1.)/2. )
            # Normalized polynomial
            p[ncount] = ( p_i*q_i*p_j )*factor
            # Derivatives with respect to (r,s)
            dp_dr = ( (dp_i)*q_i*p_j )*factor
            dp_ds = ( p_i*(dq_i*p_j+q_i*dp_j) )*factor
            # Derivatives with respect to (xi,eta)
            dp_dxi[ncount]  = dp_dr*dr_dxi
            dp_deta[ncount] = dp_dr*dr_deta + dp_ds

            ncount += 1

    return p,dp_dxi,dp_deta



# 3D - TET
#------------------------------------------------------------------------------------------------------------------#
#------------------------------------------------------------------------------------------------------------------#
#------------------------------------------------------------------------------------------------------------------#

def NormalisedJacobi3D(C,x):

    # Computes the ortogonal base of 3D polynomials of degree less 
    # or equal to n at the point x=(r,s,t) in [-1,1]^3

    N = int((C+2)*(C+3)*(C+4)/6.)
    p = np.zeros(N)

    r = x[0]; s = x[1]; t = x[2]

    # Ordering: 1st incresing the degree and 2nd lexicogafic order
    ncount = 0
    # Loop on degree
    for nDeg in range(0,C+2):
        # Loop increasing i
        for i in range(0,nDeg+1):
            if i==0:
                p_i = 1;  q_i = 1 
            else:
                p_i = JacobiPolynomials(i,r,0.,0.)[-1];     q_i = q_i*(1.-s)/2.
            # Loop increasing j
            for j in range(0,nDeg-i+1):
                if j==0:
                    p_j = 1;  q_j = ((1.-t)/2.)**i
                else:
                    p_j = JacobiPolynomials(j,s,2.*i+1.,0.)[-1];    q_j = q_j*(1.-t)/2.
                # Value for k
                k = nDeg-(i+j)
                if k==0:
                   p_k = 1.
                else:
                    p_k = JacobiPolynomials(k,t,2.*(i+j)+2.,0.)[-1]
                # factor = np.sqrt( (2.*i+1.)*(i+j+1.)*(2.*(i+j+k)+3.)/4. )
                factor = math.sqrt( (2.*i+1.)*(i+j+1.)*(2.*(i+j+k)+3.)/4. )
                p[ncount] = ( p_i*q_i*p_j*q_j*p_k )*factor
                ncount += 1

    return p


def NormalisedJacobiTet(C,x):

    # Computes the ortogonal base of 3D polynomials of degree less 
    # or equal to n at the point x=(xi,eta,zeta) in the reference tetrahedra


    xi = x[0]; eta = x[1]; zeta = x[2]

    if (eta+zeta)==0:
        r = -1; s=1
    elif zeta==1:
        r = -1; s=1  # or s=-1 (check that nothing changes)
    else:
        r = -2.*(1+xi)/(eta+zeta)-1.;
        s = 2.*(1+eta)/(1-zeta)-1.;
    t = zeta

    return NormalisedJacobi3D(C,[r,s,t])
    # return NormalisedJacobi3D_Native(C,[r,s,t])


def GradNormalisedJacobiTet(C,x,EvalOpt=0):

    # Computes the ortogonal base of 3D polynomials of degree less 
    # or equal to n at the point x=(r,s,t) in [-1,1]^3

    N = int((C+2)*(C+3)*(C+4)/6.)
    p = np.zeros(N)

    dp_dxi   = np.zeros(N)
    dp_deta  = np.zeros(N)
    dp_dzeta = np.zeros(N)

    r = x[0]; s = x[1]; t = x[2]

    # THIS MAY RUIN THE CONVERGENCE, BUT FOR POST PROCESSING ITS FINE
    if EvalOpt==1:
        if t==1.:
            t=0.999999999999
        if np.isclose(s,1.):
            s=0.999999999999

    if np.isclose(s,1.):
            s=0.99999999999999

    eta = (1./2.)*(s-s*t-1.-t)
    xi = -(1./2.)*(r+1)*(eta+t)-1.
    zeta = 1.0*t
    
    # THIS MAY RUIN THE CONVERGENCE, BUT FOR POST PROCESSING ITS FINE
    if eta == 0. and zeta == 0.:
        eta = 1.0e-14
        zeta = 1e-14

    dr_dxi   = -2./(eta+zeta)
    dr_deta  = 2.*(1.+xi)/(eta+zeta)**2
    dr_dzeta = dr_deta

    ds_deta  = 2./(1.-zeta)
    ds_dzeta = 2.*(1.+eta)/(1.-zeta)**2

    # Derivative of t is not needed because t=zeta

    #--------------------------------------------------------
    # if np.allclose(eta+zeta,0):
    #     dr_dxi   = -2./(0.001)**2
    #     dr_deta  = 2.*(1.+xi)/(0.001)**2
    # else:
    #     dr_dxi   = -2./(eta+zeta)
    #     dr_deta  = 2.*(1.+xi)/(eta+zeta)**2
    
    # dr_dzeta = dr_deta

    # if np.allclose(eta+zeta,0):
    #     ds_deta = 2./(0.001)
    #     ds_dzeta =  2.*(1.+eta)/(0.001)**2
    # else: 
    #     ds_deta  = 2./(1.-zeta)
    #     ds_dzeta = 2.*(1.+eta)/(1.-zeta)**2
    #--------------------------------------------------------

    # Ordering: 1st increasing the degree and 2nd lexicogafic order
    ncount = 0
    # Loop on degree
    for nDeg in range(0,C+2):
        # Loop increasing i
        for i in range(0,nDeg+1):
            if i==0:
                p_i = 1.;  q_i = 1.;  dp_i = 0.; dq_i = 0.
            else:
                p_i = JacobiPolynomials(i,r,0.,0.)[-1]; dp_i = JacobiPolynomials(i-1,r,1.,1.)[-1]*(i+1.)/2.    
                q_i = q_i*(1.-s)/2.; dq_i = q_i*(-i)/(1.-s)
            # Loop increasing j
            for j in range(0,nDeg-i+1):
                if j==0:
                    p_j = 1;  q_j = ((1.-t)/2.)**i;  dp_j = 0; dq_j = q_j*(-(i+j))/(1.-t);  
                else:
                    p_j = JacobiPolynomials(j,s,2.*i+1.,0.)[-1]; dp_j = JacobiPolynomials(j-1,s,2.*i+2.,1.)[-1]*(j+2.*i+2.)/2.  
                    q_j = q_j*(1.-t)/2.;  dq_j = q_j*(-(i+j))/(1.-t)

                # Value for k
                k = nDeg-(i+j);
                if k==0:
                    p_k = 1.;  dp_k = 0.; 
                else:
                    p_k = JacobiPolynomials(k,t,2.*(i+j)+2.,0.)[-1];  dp_k = JacobiPolynomials(k-1,t,2.*(i+j)+3.,1.)[-1]*(k+2.*i+2.*j+3.)/2.
        
                # factor = np.sqrt( (2.*i+1.)*(i+j+1.)*(2.*(i+j+k)+3.)/4. )
                factor = math.sqrt( (2.*i+1.)*(i+j+1.)*(2.*(i+j+k)+3.)/4. )
                # Normalized polynomial
                p[ncount] = ( p_i*q_i*p_j*q_j*p_k )*factor
                # Derivatives with respect to (r,s,t)
                dp_dr = ( (dp_i)*q_i*p_j*q_j*p_k )*factor 
                dp_ds = ( p_i*(dq_i*p_j+q_i*dp_j)*q_j*p_k )*factor
                dp_dt = ( p_i*q_i*p_j*(dq_j*p_k+q_j*dp_k) )*factor
                # if np.isnan(dp_ds):
                    # print p_i, dq_i, p_j, q_i, dp_j, q_j, p_k, factor
                # Derivatives with respect to (xi,eta,zeta)
                dp_dxi[ncount]   = dp_dr*dr_dxi
                dp_deta[ncount]  = dp_dr*dr_deta + dp_ds*ds_deta
                dp_dzeta[ncount] = dp_dr*dr_dzeta + dp_ds*ds_dzeta + dp_dt

                ncount += 1

    return p,dp_dxi,dp_deta,dp_dzeta


import numpy as np
cimport numpy as np

ctypedef long long Integer
ctypedef unsigned long long UInteger
ctypedef double Real


cdef extern from "_LowLevelAssemblyDPF_.h" nogil:
    void _GlobalAssemblyDPF_(const Real *points, 
                            const UInteger* elements, 
                            const Real* Eulerx,
                            const Real* Eulerp,
                            const Real* bases,
                            const Real* Jm,
                            const Real* AllGauss,
                            Integer ndim,
                            Integer nvar,  
                            Integer ngauss, 
                            Integer nelem, 
                            Integer nodeperelem,
                            Integer nnode,
                            Integer H_VoigtSize,
                            Integer requires_geometry_update, 
                            Integer* local_rows_stiff,
                            Integer* local_cols_stiff,
                            int *I_stiff,
                            int *J_stiff,
                            Real *V_stiff,
                            Real *T,
                            Integer is_dynamic,
                            Integer* local_rows_mass,
                            Integer* local_cols_mass,
                            int *I_mass,
                            int *J_mass,
                            Real *V_mass,
                            Real rho,
                            Real mu,
                            Real mu1,
                            Real mu2,
                            Real mu3,
                            Real mue,
                            Real lamb,
                            Real eps_1,
                            Real eps_2,
                            Real eps_3,
                            Real eps_e,
                            const Real *anisotropic_orientations
                            )


def _LowLevelAssemblyDPF_(fem_solver, function_space, formulation, mesh, material, Real[:,::1] Eulerx, Real[::1] Eulerp):

    # GET VARIABLES FOR DISPATCHING TO C
    cdef Integer ndim                       = formulation.ndim
    cdef Integer nvar                       = formulation.nvar
    cdef Integer ngauss                     = function_space.AllGauss.shape[0]
    cdef Integer nelem                      = mesh.nelem
    cdef Integer nodeperelem                = mesh.elements.shape[1]
    cdef Integer nnode                      = mesh.points.shape[0]
    cdef Integer H_VoigtSize                = material.H_VoigtSize

    cdef np.ndarray[UInteger,ndim=2, mode='c'] elements = mesh.elements
    cdef np.ndarray[Real,ndim=2, mode='c'] points       = mesh.points
    cdef np.ndarray[Real,ndim=2, mode='c'] bases        = function_space.Bases
    cdef np.ndarray[Real,ndim=3, mode='c'] Jm           = function_space.Jm
    cdef np.ndarray[Real,ndim=1, mode='c'] AllGauss     = function_space.AllGauss.flatten()

    cdef Integer requires_geometry_update               = fem_solver.requires_geometry_update
    cdef Integer is_dynamic                             = fem_solver.analysis_type != "static" and fem_solver.is_mass_computed is False

    cdef np.ndarray[Integer,ndim=1,mode='c'] local_rows_stiffness   = formulation.local_rows
    cdef np.ndarray[Integer,ndim=1,mode='c'] local_cols_stiffness   = formulation.local_columns

    cdef np.ndarray[Integer,ndim=1,mode='c'] local_rows_mass        = formulation.local_rows_mass
    cdef np.ndarray[Integer,ndim=1,mode='c'] local_cols_mass        = formulation.local_columns_mass

    # ALLOCATE VECTORS FOR SPARSE ASSEMBLY OF STIFFNESS MATRIX - CHANGE TYPES TO INT64 FOR DoF > 1e09
    cdef np.ndarray[int,ndim=1,mode='c'] I_stiffness    = np.zeros(int((nvar*nodeperelem)**2*nelem),dtype=np.int32)
    cdef np.ndarray[int,ndim=1,mode='c'] J_stiffness    = np.zeros(int((nvar*nodeperelem)**2*nelem),dtype=np.int32)
    cdef np.ndarray[Real,ndim=1,mode='c'] V_stiffness   = np.zeros(int((nvar*nodeperelem)**2*nelem),dtype=np.float64)

    cdef np.ndarray[int,ndim=1,mode='c'] I_mass         = np.zeros(1,np.int32)
    cdef np.ndarray[int,ndim=1,mode='c'] J_mass         = np.zeros(1,np.int32)
    cdef np.ndarray[Real,ndim=1,mode='c'] V_mass        = np.zeros(1,np.float64)

    if is_dynamic:
        I_mass          = np.zeros(int((nvar*nodeperelem)**2*nelem),dtype=np.int32)
        J_mass          = np.zeros(int((nvar*nodeperelem)**2*nelem),dtype=np.int32)
        V_mass          = np.zeros(int((nvar*nodeperelem)**2*nelem),dtype=np.float64)


    cdef np.ndarray[Real,ndim=1,mode='c'] T = np.zeros(mesh.points.shape[0]*nvar,np.float64)

    cdef np.ndarray[Real,ndim=2,mode='c'] anisotropic_orientations = np.zeros((1,1),np.float64)
    if material.is_transversely_isotropic:
        anisotropic_orientations = material.anisotropic_orientations

    cdef Real mu=0.,mu1=0.,mu2=0.,mu3=0.,mue=0.,lamb=0.,eps_1=0.,eps_2=0., eps_3=0., eps_e=0.

    mu, lamb, eps_1 = material.mu, material.lamb, material.eps_1

    cdef Real rho = material.rho

    _GlobalAssemblyDPF_(    &points[0,0],
                            &elements[0,0],
                            &Eulerx[0,0],
                            &Eulerp[0],
                            &bases[0,0],
                            &Jm[0,0,0],
                            &AllGauss[0],
                            ndim,
                            nvar,
                            ngauss,
                            nelem,
                            nodeperelem,
                            nnode,
                            H_VoigtSize,
                            requires_geometry_update,
                            &local_rows_stiffness[0],
                            &local_cols_stiffness[0],
                            &I_stiffness[0],
                            &J_stiffness[0],
                            &V_stiffness[0],
                            &T[0],
                            is_dynamic,
                            &local_rows_mass[0],
                            &local_cols_mass[0],
                            &I_mass[0],
                            &J_mass[0],
                            &V_mass[0],
                            rho,
                            mu,
                            mu1,
                            mu2,
                            mu3,
                            mue,
                            lamb,
                            eps_1,
                            eps_2,
                            eps_3,
                            eps_e,
                            &anisotropic_orientations[0,0]
                            )


    return I_stiffness, J_stiffness, V_stiffness, I_mass, J_mass, V_mass, T






import numpy as np
from numpy import einsum
from .MaterialBase import Material
from Florence.Tensor import trace, Voigt


class MooneyRivlin_1(Material):
    """ Polyconvex compressible MooneyRivlin material model based on the energy:

            W = alpha*C:I+beta*G:I+lambda/2*(J-1)**2-4*beta*J-2*alpha*lnJ - (3*alpha-beta)

        where at the origin (alpha + beta) = mu/2
        """

    def __init__(self, ndim, **kwargs):
        mtype = type(self).__name__
        super(MooneyRivlin_1, self).__init__(mtype,ndim,**kwargs)
        
        self.alpha = self.mu/4.0
        self.beta = self.mu/4.0

        self.is_transversely_isotropic = False
        self.energy_type = "internal_energy"
        self.nature = "nonlinear"
        self.fields = "mechanics"

        if self.ndim==3:
            self.H_VoigtSize = 6
        elif self.ndim==2:
            self.H_VoigtSize = 3

        # LOW LEVEL DISPATCHER
        self.has_low_level_dispatcher = False


    def Hessian(self,StrainTensors,ElectricFieldx=0,elem=0,gcounter=0):

        lamb = self.lamb

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]

        # Actual version
        # d = np.einsum
        # H_Voigt = Voigt( 4.0*beta/J*d('ij,kl',b,b) - 2.0*beta/J*( d('ik,jl',b,b) + d('il,jk',b,b) ) +\
        #   (lamb+4.0*beta+4.0*alpha/J)*d('ij,kl',I,I) + 2.0*(lamb*(J-1.0) -4.0*beta -2.0*alpha/J)*d('ij,kl',I,I) -\
        #   1.0*(lamb*(J-1.0) -4.0*beta -2.0*alpha/J)*(d('ik,jl',I,I)+d('il,jk',I,I)) ,1) 

        # # Simplified version
        # H_Voigt = 2.0*beta/J*( 2.0*einsum('ij,kl',b,b) - einsum('ik,jl',b,b) - einsum('il,jk',b,b) ) + \
        #   (lamb*(2.0*J-1.0) -4.0*beta)*einsum('ij,kl',I,I) - \
        #   (lamb*(J-1.0) -4.0*beta -2.0*alpha/J)*( einsum('ik,jl',I,I) + einsum('il,jk',I,I) )

        # Further simplified version
        H_Voigt = 2.0*self.beta/J*( 2.0*einsum('ij,kl',b,b) - einsum('ik,jl',b,b) - einsum('il,jk',b,b) ) + \
            (lamb*(2.0*J-1.0) -4.0*self.beta)*self.Iijkl - \
            (lamb*(J-1.0) -4.0*self.beta -2.0*self.alpha/J)*self.Iikjl

        H_Voigt = Voigt(H_Voigt,1) 


        self.H_VoigtSize = H_Voigt.shape[0]

        return H_Voigt



    def CauchyStress(self,StrainTensors,ElectricFieldx,elem=0,gcounter=0):

        lamb = self.lamb

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]


        if self.ndim == 3:
            trb = trace(b)
        elif self.ndim == 2:
            trb = trace(b) + 1

        # stress = 2.0*alpha/J*b+2.0*beta/J*(trace(b)*b - np.dot(b,b)) + (lamb*(J-1.0)-4.0*beta-2.0*alpha/J)*I 
        stress = 2.0*self.alpha/J*b+2.0*self.beta/J*(trb*b - np.dot(b,b)) + (lamb*(J-1.0)-4.0*self.beta-2.0*self.alpha/J)*I 
        # print stress
        return stress


    def ElectricDisplacementx(self,StrainTensors,ElectricFieldx):
        ndim = StrainTensors['I'].shape[0]
        return np.zeros((ndim,1))



import numpy as np
from .LineBP import LagrangeBP_, LagrangeGaussLobattoBP_
try:
    from functools import lru_cache
except ImportError:
    from Florence.Utils.backports import lru_cache

# def Lagrange(C,xi):
#     n = C+2
#     nsize = n-1
#     ndiv = 2.0/nsize
#     eps = 1.0*np.zeros(n)
#     eps[0]=-1.; eps[n-1]=1.

#     for i in range(0,nsize):
#         eps[i+1] = eps[i]+ndiv

#     A = 1.0*np.zeros((n,n))
#     A[:,0] = np.ones(n)

#     for i in range(1,n):
#         for j in range(0,n):
#             A[j,i] = pow(eps[j],i)


#     N = 1.0*np.zeros(n); dN=1.0*np.zeros(n)

#     for ishape in range(0,n):
#         RHS = 1.0*np.zeros(n)
#         RHS[ishape] = 1.

#         # Solve linear system (dense LU)
#         coeff = np.linalg.solve(A,RHS)
#         # Build shape functions 
#         for incr in range(0,n):
#             N[ishape] = N[ishape]+coeff[incr]*pow(xi,incr)

#         # Build derivate of shape functions
#         for incr in range(0,n-1):
#             dN[ishape] = dN[ishape]+(incr+1)*coeff[incr+1]*pow(xi,incr)


#     return (N,dN,eps) 



@lru_cache(maxsize=None)
def Lagrange(C,xi):

    n = C+2
    ranger = np.arange(n)
    eps = np.linspace(-1.,1.,n)

    A = np.zeros((n,n))
    A[:,0] = 1.
    for i in range(1,n):
        A[:,i] = eps**i

    RHS = np.zeros((n,n))
    np.fill_diagonal(RHS,1)
    coeff = np.linalg.solve(A,RHS)
    xis = np.ones(n)*xi**ranger
    N = np.dot(coeff.T,xis)
    dN = np.dot(coeff[1:,:].T,xis[:-1]*(1+ranger[:-1]))

    return (N,dN,eps) 


# def LagrangeGaussLobatto(C,xi):
    
#     from Florence.QuadratureRules import GaussLobattoQuadrature

#     n = C+2
#     nsize = n-1
#     ndiv = 2.0/nsize

#     eps = GaussLobattoQuadrature(n)[0]

#     A = 1.0*np.zeros((n,n))
#     A[:,0] = np.ones(n)

#     for i in range(1,n):
#         for j in range(0,n):
#             A[j,i] = pow(eps[j],i)


#     N = 1.0*np.zeros(n); dN=1.0*np.zeros(n)

#     for ishape in range(0,n):
#         RHS = 1.0*np.zeros(n)
#         RHS[ishape] = 1.

#         # Solve linear system (dense LU)
#         coeff = np.linalg.solve(A,RHS)
#         # Build shape functions 
#         for incr in range(0,n):
#             N[ishape] = N[ishape]+coeff[incr]*pow(xi,incr)

#         # Build derivate of shape functions
#         for incr in range(0,n-1):
#             dN[ishape] = dN[ishape]+(incr+1)*coeff[incr+1]*pow(xi,incr)


#     return (N,dN,eps)


@lru_cache(maxsize=None)
def LagrangeGaussLobatto(C,xi):
    
    from Florence.QuadratureRules import GaussLobattoQuadrature

    n = C+2
    ranger = np.arange(n)

    eps = GaussLobattoQuadrature(n)[0][:,0]

    A = np.zeros((n,n))
    A[:,0] = 1.
    for i in range(1,n):
        A[:,i] = eps**i
    # A1[:,1:] = np.array([eps**i for i in range(1,n)]).T[0,:,:]


    RHS = np.zeros((n,n))
    np.fill_diagonal(RHS,1)
    coeff = np.linalg.solve(A,RHS)
    xis = np.ones(n)*xi**ranger
    N = np.dot(coeff.T,xis)
    # dN = np.einsum('i,ij,i',1+ranger[:-1],coeff[1:,:],xis[:-1])
    dN = np.dot(coeff[1:,:].T,xis[:-1]*(1+ranger[:-1]))

    return (N,dN,eps)




@lru_cache(maxsize=None)
def Legendre(C,xi):
    # For Linear Basis Generating Legendre Polynomials is Not Required
    if C==0:
        N = np.array([1.0/2*(1-xi), 1.0/2*(1+xi)])
        dN = np.array([-1.0/2, 1./2])

    # For Higher Order 
    elif C>0:
        # The first two Legendre polynomials 
        p0 = 1.0; p1 = xi
        # Derivatives of The First Two Legendre Polynomials 
        dp0 = 0.0; dp1 = 1.0
        # Allocate size and dimensions
        ndim = C+2
        P = np.zeros((ndim+1,1)); dP = np.zeros((ndim+1,1))
        N = np.zeros((ndim+1,1)); dN = np.zeros((ndim+1,1))
        P[0] = p0; P[1] = p1
        dP[0] = dp0; dP[1] = dp1
        # Generate Legendre Polynomials
        for i in range(2,ndim+1):
            P[i]  = ((2.0*i-1)*xi*P[i-1] - (i-1)*P[i-2])/(i)
            dP[i]  = ((2.0*i-1)*xi*dP[i-1] + (2.0*i-1)*P[i-1] - (i-1)*dP[i-2])/(i)

        # From Legendre Polynomials Generate FE Basis Functions 
        for i in range(3,ndim+2):
            # N[i-1] =  (P[i-1]-P[i-3])/np.sqrt(2*(2*i-3))
            # dN[i-1] =  (dP[i-1]-dP[i-3])/np.sqrt(2*(2*i-3))
            # Ledger's Normalisation 
            N[i-1] =  (P[i-1]-P[i-3])/((2.0*i-3.))
            dN[i-1] =  (dP[i-1]-dP[i-3])/((2.0*i-3.))


        # Put the hat functions at exterior nodes  
        N = np.append([np.append([1.0/2.0*(1.0-xi)],N[2:-1])],[1.0/2*(1.0+xi)])
        dN = np.append([np.append([-0.5],dN[2:-1])],[0.5])


    return (N,dN)










# def LagrangeBP(C,xi):

#     n = C+2
#     eps = np.linspace(-1.,1.,n)

#     N = 1.0*np.zeros(n); dN=1.0*np.zeros(n)

#     for ishape in range(0,n):

#         d = np.zeros(n)
#         d[ishape] = 1.
        
#         # Find the Newton Divided Difference
#         for k in range(0,n):
#             for j in reversed(range(k+1,n)):
#                 d[j] = (d[j]-d[j-1])/(eps[j]-eps[j-k-1])

#         # Convert to Monomials
#         for k in reversed(range(0,n)):
#             for j in range(k,n-1):
#                 d[j] -= eps[k]*d[j+1]

        
#         # Build shape functions 
#         for incr in range(0,n):
#             N[ishape] += d[incr]*pow(xi,incr)

#         # Build derivative of shape functions
#         for incr in range(0,n-1):
#             dN[ishape] += (incr+1)*d[incr+1]*pow(xi,incr)


#     return (N,dN,eps) 



# Bjorck Peryra Bases
# This technique blows faster than Vandermonde matrices specially
# beyond C=24

@lru_cache(maxsize=None)
def LagrangeBP(C,xi):
    return LagrangeBP_(C,xi)

@lru_cache(maxsize=None)
def LagrangeGaussLobattoBP(C,xi):
    
    n = C+2
    from Florence.QuadratureRules import GaussLobattoQuadrature
    eps = GaussLobattoQuadrature(n)[0][:,0].copy()
    return LagrangeGaussLobattoBP_(C,xi,eps)


import numpy as np
from numpy import einsum
from .MaterialBase import Material
from Florence.Tensor import trace, Voigt

__all__ = ["CoupleStressModel"]

class CoupleStressModel(Material):
    """The couple stress model based on linear strain and curvature tensors

        W_cs(e,x) = W_lin(e) + eta*x:x

    """

    def __init__(self, ndim, **kwargs):
        mtype = type(self).__name__
        super(CoupleStressModel, self).__init__(mtype, ndim, **kwargs)
        # REQUIRES SEPARATELY
        self.energy_type = "internal_energy"
        self.nature = "linear"
        self.fields = "couple_stress"

        # FOR STATICALLY CONDENSED FORMULATION
        if self.ndim==3:
            self.elasticity_tensor_size = 6
            self.H_VoigtSize = 6
        elif self.ndim==2:
            self.elasticity_tensor_size = 3
            self.H_VoigtSize = 3
        self.gradient_elasticity_tensor_size = ndim

        # LOW LEVEL DISPATCHER
        # self.has_low_level_dispatcher = True
        self.has_low_level_dispatcher = False

    def KineticMeasures(self,F,ElectricFieldx=0, elem=0):
        # raise NotImplementedError("KineticMeasures for this formulation is not implemented yet")
        return None, self.elasticity_tensors, self.gradient_elasticity_tensors


    def Hessian(self,StrainTensors,ElectricFieldx,elem=0,gcounter=0):

        mu = self.mu
        lamb = self.lamb
        eta = self.eta
        I = StrainTensors['I']

        self.elasticity_tensor = Voigt(lamb*einsum('ij,kl',I,I)+mu*(einsum('ik,jl',I,I)+einsum('il,jk',I,I)),1)
        self.gradient_elasticity_tensor = 2.*eta*I
        self.coupling_tensor0 = np.zeros((self.elasticity_tensor.shape[0],self.gradient_elasticity_tensor.shape[0]))

        # # BUILD HESSIAN
        # factor = 1.
        # H1 = np.concatenate((self.elasticity_tensor,factor*self.coupling_tensor0),axis=1)
        # H2 = np.concatenate((factor*self.coupling_tensor0.T,self.gradient_elasticity_tensor),axis=1)
        # H_Voigt = np.concatenate((H1,H2),axis=0)
        # return H_Voigt
        return None



    def CauchyStress(self,StrainTensors,ElectricFieldx,elem=0,gcounter=0):

        mu = self.mu
        lamb = self.lamb
        eta = self.eta
        I = StrainTensors['I']
        e = StrainTensors['strain'][gcounter]

        tre = trace(e)
        if self.ndim==2:
            tre +=1.

        sigma = lamb*tre*I + 2.0*mu*e
        return sigma


    def CoupleStress(self,StrainTensors,ElectricFieldx,elem=0,gcounter=0):

        mu = self.mu
        lamb = self.lamb
        eta = self.eta
        I = StrainTensors['I']
        F = StrainTensors['F'][gcounter]
        GradU = F - I
        k = GradU - GradU.T
        k = np.array([k[0,1],k[0,2],k[1,2]])

        couple_stress = 2*eta*k
        return couple_stress


    def LagrangeMultiplierStress(self,S,ElectricFieldx=None,elem=0,gcounter=0):

        lm_stress = 0.5*S
        return lm_stress


    def TotalStress(self,StrainTensors,S,ElectricFieldx=None,elem=0,gcounter=0):

        cauchy_stress = self.CauchyStress(StrainTensors,ElectricFieldx,elem=elem,gcounter=gcounter)
        lm_stress     = self.LagrangeMultiplierStress(S,ElectricFieldx,elem=elem,gcounter=gcounter)
        lm_stress     = np.array([
            [0.,lm_stress[0],lm_stress[1]],
            [-lm_stress[0],0.,lm_stress[2]],
            [-lm_stress[1],-lm_stress[2],0.]
            ])
        return cauchy_stress + lm_stress



#cython: profile=False
#cython: infer_types=True
#cython: boundscheck=False
#cython: wraparound=False
#cython: nonecheck=False

import numpy as np
cimport numpy as np

ctypedef double Real


cdef extern from "_NeoHookean_.h" nogil:
    cdef cppclass _NeoHookean_[Real]:
        _NeoHookean_() except +
        _NeoHookean_(Real mu, Real lamb) except +
        void SetParameters(Real mu, Real lamb) except +
        void KineticMeasures(Real *Snp, Real* Hnp, int ndim, int ngauss, const Real *Fnp) except +



def KineticMeasures(material, np.ndarray[Real, ndim=3, mode='c'] F):
    
    cdef int ndim = F.shape[2]
    cdef int ngauss = F.shape[0]
    cdef np.ndarray[Real, ndim=3, mode='c'] stress, hessian

    stress = np.zeros((ngauss,ndim,ndim),dtype=np.float64)
    if ndim==3:
        hessian = np.zeros((ngauss,6,6),dtype=np.float64)
    elif ndim==2:
        hessian = np.zeros((ngauss,3,3),dtype=np.float64)

    cdef _NeoHookean_[Real] mat_obj = _NeoHookean_()
    mat_obj.SetParameters(material.mu,material.lamb)
    mat_obj.KineticMeasures(&stress[0,0,0], &hessian[0,0,0], ndim, ngauss, &F[0,0,0])

    return stress, hessian


from warnings import warn
import numpy as np
from scipy.sparse import csr_matrix, csc_matrix

try:
    from ._LowLevelAssemblyDF__LinearElastic_ import _LowLevelAssemblyDF__LinearElastic_
    from ._LowLevelAssemblyDF__LinearElastic_ import _LowLevelAssemblyDF__LinearElastic_ as _LowLevelAssemblyDF__IncrementalLinearElastic_
    from ._LowLevelAssemblyDF__NeoHookean_ import _LowLevelAssemblyDF__NeoHookean_
    from ._LowLevelAssemblyDF__MooneyRivlin_ import _LowLevelAssemblyDF__MooneyRivlin_
    from ._LowLevelAssemblyDF__AnisotropicMooneyRivlin_1_ import _LowLevelAssemblyDF__AnisotropicMooneyRivlin_1_
    from ._LowLevelAssemblyDF__NearlyIncompressibleMooneyRivlin_ import _LowLevelAssemblyDF__NearlyIncompressibleMooneyRivlin_
    from ._LowLevelAssemblyDPF__IsotropicElectroMechanics_0_ import _LowLevelAssemblyDPF__IsotropicElectroMechanics_0_
    from ._LowLevelAssemblyDPF__IsotropicElectroMechanics_3_ import _LowLevelAssemblyDPF__IsotropicElectroMechanics_3_
    from ._LowLevelAssemblyDPF__SteinmannModel_ import _LowLevelAssemblyDPF__SteinmannModel_
    from ._LowLevelAssemblyDPF__IsotropicElectroMechanics_101_ import _LowLevelAssemblyDPF__IsotropicElectroMechanics_101_
    from ._LowLevelAssemblyDPF__IsotropicElectroMechanics_105_ import _LowLevelAssemblyDPF__IsotropicElectroMechanics_105_
    from ._LowLevelAssemblyDPF__IsotropicElectroMechanics_106_ import _LowLevelAssemblyDPF__IsotropicElectroMechanics_106_
    from ._LowLevelAssemblyDPF__IsotropicElectroMechanics_107_ import _LowLevelAssemblyDPF__IsotropicElectroMechanics_107_
    from ._LowLevelAssemblyDPF__IsotropicElectroMechanics_108_ import _LowLevelAssemblyDPF__IsotropicElectroMechanics_108_
    from ._LowLevelAssemblyDPF__Piezoelectric_100_ import _LowLevelAssemblyDPF__Piezoelectric_100_
    has_low_level_dispatcher = True
except ImportError:
    has_low_level_dispatcher = False
    warn("Cannot use low level dispatchers for Assembly")


try:
    from ._LowLevelAssemblyExplicit_DF_DPF_ import _LowLevelAssemblyExplicit_DF_DPF_
    has_low_level_dispatcher = True
except ImportError:
    has_low_level_dispatcher = False
    warn("Cannot use low level dispatchers for Assembly")

try:
    from ._LowLevelAssemblyPerfectLaplacian_ import _LowLevelAssemblyPerfectLaplacian_
    has_low_level_dispatcher = True
except ImportError:
    has_low_level_dispatcher = False
    warn("Cannot use low level dispatchers for Assembly")


__all__ = ['_LowLevelAssembly_', '_LowLevelAssemblyExplicit_', '_LowLevelAssemblyLaplacian_']


def _LowLevelAssembly_(fem_solver, function_space, formulation, mesh, material, Eulerx, Eulerp):

    prefix = "_LowLevelAssemblyDF__"
    if formulation.fields == "electro_mechanics":
        prefix = "_LowLevelAssemblyDPF__"

    assembly_func = prefix + type(material).__name__ + "_"
    # CHECK IF LOW LEVEL ASSEMBLY EXISTS FOR MATERIAL
    ll_exists = False
    for key in globals().keys():
        if assembly_func == key:
            ll_exists = True
            break
    if ll_exists is False:
        raise NotImplementedError("Turning optimise option on for {} material is not supported yet. Consider 'optimise=False' for now".format(type(material).__name__))


    # MAKE MESH DATA CONTIGUOUS
    mesh.ChangeType()

    # CALL LOW LEVEL ASSEMBLER
    I_stiffness, J_stiffness, V_stiffness, I_mass, J_mass, V_mass, \
        T = eval(assembly_func)(fem_solver, function_space, formulation, mesh, material, Eulerx, Eulerp)

    nvar = formulation.nvar
    stiffness = csr_matrix((V_stiffness,(I_stiffness,J_stiffness)),
        shape=((nvar*mesh.points.shape[0],nvar*mesh.points.shape[0])),dtype=np.float64)

    F, mass = [], []

    if fem_solver.analysis_type != "static" and fem_solver.is_mass_computed is False:
        mass = csr_matrix((V_mass,(I_mass,J_mass)),
            shape=((nvar*mesh.points.shape[0],nvar*mesh.points.shape[0])),dtype=np.float64)

    return stiffness, T, F, mass



def _LowLevelAssembly_Par_(fem_solver, function_space, formulation, mesh, material, Eulerx, Eulerp):

    prefix = "_LowLevelAssemblyDF__"
    if formulation.fields == "electro_mechanics":
        prefix = "_LowLevelAssemblyDPF__"
        
    assembly_func = prefix + type(material).__name__ + "_"
    # CHECK IF LOW LEVEL ASSEMBLY EXISTS FOR MATERIAL
    ll_exists = False
    for key in globals().keys():
        if assembly_func == key:
            ll_exists = True
            break
    if ll_exists is False:
        raise NotImplementedError("Turning optimise option on for {} material is not supported yet. Consider 'optimise=False' for now".format(type(material).__name__))

    # CALL LOW LEVEL ASSEMBLER
    return eval(assembly_func)(fem_solver, function_space, formulation, mesh, material, Eulerx, Eulerp)



def _LowLevelAssemblyExplicit_(fem_solver, function_space, formulation, mesh, material, Eulerx, Eulerp):

    # MAKE MESH DATA CONTIGUOUS
    mesh.ChangeType()

    # CALL LOW LEVEL ASSEMBLER
    I_mass, J_mass, V_mass, \
        T = _LowLevelAssemblyExplicit_DF_DPF_(function_space, formulation, mesh, material, Eulerx, Eulerp)

    F, mass = [], []

    if fem_solver.analysis_type != "static" and fem_solver.is_mass_computed is False:
        nvar = formulation.nvar
        mass = csr_matrix((V_mass,(I_mass,J_mass)),
            shape=((nvar*mesh.points.shape[0],nvar*mesh.points.shape[0])),dtype=np.float64)

    return T, F, mass


def _LowLevelAssemblyExplicit_Par_(function_space, formulation, mesh, material, Eulerx, Eulerp):
    # CALL LOW LEVEL ASSEMBLER
    return _LowLevelAssemblyExplicit_DF_DPF_(function_space, formulation, mesh, material, Eulerx, Eulerp)[-1]



def _LowLevelAssemblyLaplacian_(fem_solver, function_space, formulation, mesh, material, Eulerx, Eulerp):

    mesh.GetNumberOfNodes()
    # MAKE MESH DATA CONTIGUOUS
    mesh.ChangeType()

    # CALL LOW LEVEL ASSEMBLER
    I, J, V = _LowLevelAssemblyPerfectLaplacian_(fem_solver, function_space, formulation, mesh, material, Eulerx, Eulerp)
    stiffness = csr_matrix((V,(I,J)), shape=((mesh.nnode,mesh.nnode)),dtype=np.float64)

    return stiffness, np.zeros(mesh.nnode,np.float64)






#!/usr/bin/env python
#   Copyright 2014 Sergio Oller <sergioller@gmail.com>
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#   links:      https://pypi.python.org/pypi/parmap/1.2.0
#               http://stackoverflow.com/questions/5442910/python-multiprocessing-pool-map-for-multiple-arguments

"""
This module implements map and starmap functions (see python standard
library to learn about them).

The implementations provided in this module allow providing additional
arguments to the mapped functions. Additionally they will initialize
the pool and close it automatically by default if possible.

Use these module in CPU intensive map functions.

Usage:
    import parmap
    # You want to do:
    y = [myfunction(x, argument1, argument2) for x in mylist]
    # In parallel:
    y = parmap.map(myfunction, mylist, argument1, argument2)

    # You want to do:
    z = [myfunction(x, y, argument1, argument2) for (x,y) in mylist]
    # In parallel:
    z = parmap.starmap(myfunction, mylist, argument1, argument2)

    # Yoy want to do:
    listx = [1, 2, 3, 4, 5, 6]
    listy = [2, 3, 4, 5, 6, 7]
    param = 3.14
    param2 = 42
    listz = []
    for x in listx:
        for y in listy:
            listz.append(myfunction(x, y, param1, param2))
    # In parallel:
    listz = parmap.starmap(myfunction, zip(listx,listy), param1, param2)

"""
# The original idea for this implementation was given by J.F. Sebastian
# at  http://stackoverflow.com/a/5443941/446149


from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division

try:
    from itertools import izip
except ImportError:  # Python 3 built-in zip already returns iterable
    izip = zip

from itertools import repeat

try:
    import multiprocessing    # ORIGINAL parmap  
    # import pathos.multiprocessing as multiprocessing
    HAVE_PARALLEL = True
except ImportError:
    HAVE_PARALLEL = False

def _func_star_single(func_item_args):
    """Equivalent to:
       func = func_item_args[0]
       item = func_item_args[1]
       args = func_item_args[2:]
       return func(item,args[0],args[1],...)  
    """
    return func_item_args[0](*[func_item_args[1]] + func_item_args[2])

def _func_star_many(func_items_args):
    """Equivalent to:
       func = func_item_args[0]
       items = func_item_args[1]
       args = func_item_args[2:]
       return func(items[0],items[1],...,args[0],args[1],...)  
    """
    return func_items_args[0](*list(func_items_args[1]) + func_items_args[2])



def map(function, iterable, *args, **kwargs):
    """
    Equivalent to:
    return [function(x, args[0], args[1],...) for x in iterable]

    Keyword arguments:
       - parallel=True/False: Force parallelization on/off
       - chunksize=int: see multiprocessing.Pool().map
       - pool=multiprocessing.Pool() Pass an existing pool.
       - processes=int: see multiprocessing.Pool() processes argument
    """
    parallel = kwargs.get("parallel", HAVE_PARALLEL)
    chunksize = kwargs.get("chunksize", None)
    pool = kwargs.get("pool", None)
    processes = kwargs.get("processes", None)
    # processes = kwargs.get("ncpus", None)

    # Check if parallel is inconsistent with HAVE_PARALLEL:
    if HAVE_PARALLEL == False and parallel == True:
        print("W: Parallelization is disabled because",
              "multiprocessing is missing")
        parallel = False
    # Assign right number of processes
    if processes is None:
        processes = multiprocessing.cpu_count()
    # Initialize pool if parallel:
    if parallel and pool is None:
        try:
            pool = multiprocessing.Pool(processes=processes)
            # pool = multiprocessing.ProcessingPool(ncpus=processes)
        except AssertionError:  # Disable parallel on error:
            print("W: Could not create multiprocessing.Pool.",
                  "Parallel disabled")
            parallel = False   

    # Map:
    if parallel:
        output = pool.map(_func_star_single,
                          izip(repeat(function), iterable,repeat(list(args))))
    else:
        output = [function(*([item] + list(args))) for item in iterable]
    return output

def starmap(function, iterables, *args, **kwargs):
    """ Equivalent to:
        return [function(x1,x2,x3,..., args[0], args[1],...) for \
(x1,x2,x3...) in iterable]

    Keyword arguments:
       - parallel=True/False: Force parallelization on/off
       - chunksize=int: see multiprocessing.Pool().map
       - pool=multiprocessing.Pool() Pass an existing pool.
       - processes=int: see multiprocessing.Pool() processes argument
    """
    parallel = kwargs.get("parallel", HAVE_PARALLEL)
    chunksize = kwargs.get("chunksize", None)
    pool = kwargs.get("pool", None)
    processes = kwargs.get("processes", None)
    # Check if parallel is inconsistent with HAVE_PARALLEL:
    if HAVE_PARALLEL == False and parallel == True:
        print("W: Parallelization is disabled because",
              "multiprocessing is missing")
        parallel = False
    # Initialize pool if parallel:
    if parallel and pool is None:
        try:
            pool = multiprocessing.Pool(processes=processes)
        except AssertionError:  # Disable parallel on error:
            print("W: Could not create multiprocessing.Pool.",
                  "Parallel disabled")
            parallel = False
    # Map:
    if parallel:
        output = pool.map(_func_star_many,
                          izip(repeat(function),
                               iterables, repeat(list(args))),
                          chunksize)
        pool.close()
        pool.join()
    else:
        output = [function(*(list(item) + list(args))) for item in iterables]
    return output


if __name__ == "__main__":
    multiprocessing.freeze_support()
    def _func(*aaa):
        """ Prints and returns the inputs. Trivial example."""
        print(aaa)
        return aaa
    print("Example1: Begins")
    ITEMS = [1, 2, 3, 4]
    OUT = map(_func, ITEMS, 5, 6, 7, 8, parallel=False)
    print("Using parallel:")
    OUT_P = map(_func, ITEMS, 5, 6, 7, 8, parallel=True)
    if OUT != OUT_P:
        print("Example1: Failed")
    else:
        print("Example1: Success")
    print("Example2: Begins")
    ITEMS = [(1, 'a'), (2, 'b'), (3, 'c')]
    OUT = starmap(_func, ITEMS, 5, 6, 7, 8, parallel=False)
    print("Using parallel")
    OUT_P = starmap(_func, ITEMS, 5, 6, 7, 8, parallel=True)
    if OUT != OUT_P:
        print("Example2: Failed")
    else:
        print("Example2: Success")




#cython: profile=False
#cython: infer_types=True
#cython: boundscheck=False
#cython: wraparound=False
#cython: nonecheck=False

import numpy as np
cimport numpy as np

ctypedef double Real


cdef extern from "_ExplicitIsotropicElectroMechanics_108_.h" nogil:
    cdef cppclass _ExplicitIsotropicElectroMechanics_108_[Real]:
        _ExplicitIsotropicElectroMechanics_108_() except +
        _ExplicitIsotropicElectroMechanics_108_(Real mu1, Real mu2, Real lamb, Real eps_2) except +
        void SetParameters(Real mu1, Real mu2, Real lamb, Real eps_2) except +
        void KineticMeasures(Real *Dnp, Real *Snp, int ndim, int ngauss, const Real *Fnp, const Real *Enp) except +



def KineticMeasures(material, np.ndarray[Real, ndim=3, mode='c'] F, np.ndarray[Real, ndim=2] E):

    cdef int ndim = F.shape[2]
    cdef int ngauss = F.shape[0]
    cdef np.ndarray[Real, ndim=3, mode='c'] D, stress

    D = np.zeros((ngauss,ndim,1),dtype=np.float64)
    stress = np.zeros((ngauss,ndim,ndim),dtype=np.float64)

    cdef _ExplicitIsotropicElectroMechanics_108_[Real] mat_obj = _ExplicitIsotropicElectroMechanics_108_()
    mat_obj.SetParameters(material.mu1,material.mu2,material.lamb,material.eps_2)
    mat_obj.KineticMeasures(&D[0,0,0], &stress[0,0,0], ndim, ngauss, &F[0,0,0], &E[0,0])

    return D, stress


from __future__ import print_function
import gc, os, sys, inspect
from copy import deepcopy
from warnings import warn
from time import time
import numpy as np
import scipy as sp
from numpy.linalg import norm
from Florence.Utils import insensitive
from Florence.Tensor import trace, Voigt

from Florence.PostProcessing import *
from Florence.Solver import LinearSolver, FEMSolver

__all__ = ["CoupleStressSolver"]

class CoupleStressSolver(FEMSolver):

    def __init__(self,**kwargs):
        if 'static_condensation' in kwargs.keys():
            self.static_condensation = kwargs['static_condensation']
        else:
            self.static_condensation = True

        if 'assembly_print_counter' in kwargs.keys():
            self.assembly_print_counter = kwargs['assembly_print_counter']
            del kwargs['assembly_print_counter']
        else:
            self.assembly_print_counter = 500

        if 'lump_rhs' in kwargs.keys():
            self.lump_rhs = kwargs['lump_rhs']
            del kwargs['lump_rhs']
        else:
            self.lump_rhs = False


        super(CoupleStressSolver, self).__init__(**kwargs)
        self.gamma   = 0.5
        self.beta    = 0.25

    def GetBoundaryInfo(self, mesh, formulation, boundary_condition):

        all_dofs = np.arange(mesh.points.shape[0]*formulation.nvar)
        if formulation.fields == "electro_mechanics" or formulation.fields == "flexoelectric":
            self.electric_dofs = all_dofs[formulation.nvar-1::formulation.nvar]
            self.mechanical_dofs = np.setdiff1d(all_dofs,self.electric_dofs)

            # GET BOUNDARY CONDITON FOR THE REDUCED MECHANICAL SYSTEM
            self.columns_in_mech = np.intersect1d(boundary_condition.columns_in,self.mechanical_dofs)
            self.columns_in_mech_idx = np.in1d(self.mechanical_dofs,boundary_condition.columns_in)

            # GET BOUNDARY CONDITON FOR THE REDUCED ELECTROSTATIC SYSTEM
            self.columns_in_electric = np.intersect1d(boundary_condition.columns_in,self.electric_dofs)
            self.columns_in_electric_idx = np.in1d(self.electric_dofs,boundary_condition.columns_in)


            # GET FREE MECHANICAL DOFs
            self.columns_out_mech = np.intersect1d(boundary_condition.columns_out,self.mechanical_dofs)
            self.columns_out_mech_idx = np.in1d(self.mechanical_dofs,boundary_condition.columns_out)

            # GET FREE ELECTROSTATIC DOFs
            self.columns_out_electric = np.intersect1d(boundary_condition.columns_out,self.electric_dofs)
            self.columns_out_electric_idx = np.in1d(self.electric_dofs,boundary_condition.columns_out)

            self.applied_dirichlet_mech = boundary_condition.applied_dirichlet[np.in1d(boundary_condition.columns_out,self.columns_out_mech)]
            self.applied_dirichlet_electric = boundary_condition.applied_dirichlet[np.in1d(boundary_condition.columns_out,self.columns_out_electric)]

            # MAPPED QUANTITIES
            out_idx = np.in1d(all_dofs,boundary_condition.columns_out)
            idx_electric = all_dofs[formulation.nvar-1::formulation.nvar]
            idx_mech = np.setdiff1d(all_dofs,idx_electric)

            self.all_electric_dofs = np.arange(mesh.points.shape[0])
            self.electric_out = self.all_electric_dofs[out_idx[idx_electric]]
            self.electric_in = np.setdiff1d(self.all_electric_dofs,self.electric_out)

            self.all_mech_dofs = np.arange(mesh.points.shape[0]*formulation.ndim)
            self.mech_out = self.all_mech_dofs[out_idx[idx_mech]]
            self.mech_in = np.setdiff1d(self.all_mech_dofs,self.mech_out)

            # LOCAL NUMBERING OF FIELDS
            self.all_local_dofs = np.arange(mesh.elements.shape[1]*formulation.nvar)
            self.all_local_electric_dofs = self.all_local_dofs[formulation.nvar-1::formulation.nvar]
            self.all_local_mech_dofs = np.setdiff1d(self.all_local_dofs,self.all_local_electric_dofs)

            # INTERMIX RAVELLED LOCAL INDICES
            matrix_shape = (mesh.elements.shape[1]*formulation.nvar,mesh.elements.shape[1]*formulation.nvar)
            self.idx_uu = np.ravel_multi_index(np.meshgrid(self.all_local_mech_dofs,self.all_local_mech_dofs), matrix_shape)
            self.idx_up = np.ravel_multi_index(np.meshgrid(self.all_local_mech_dofs,self.all_local_electric_dofs), matrix_shape)
            self.idx_pu = np.ravel_multi_index(np.meshgrid(self.all_local_electric_dofs,self.all_local_mech_dofs), matrix_shape)
            self.idx_pp = np.ravel_multi_index(np.meshgrid(self.all_local_electric_dofs,self.all_local_electric_dofs), matrix_shape)


        elif formulation.fields == "mechanics" or formulation.fields == "couple_stress":
            self.electric_dofs = []
            self.mechanical_dofs = all_dofs
            self.columns_out_mech = boundary_condition.columns_out

            self.mech_in = boundary_condition.columns_in
            self.mech_out = boundary_condition.columns_out

            self.applied_dirichlet_mech = boundary_condition.applied_dirichlet



    def __checkdata__(self, material, boundary_condition,
        formulation, mesh, function_spaces, solver, contact_formulation=None):

        ## IMPORTANT
        self.requires_geometry_update = False
        self.analysis_nature = "linear"

        function_spaces, solver = super(CoupleStressSolver,self).__checkdata__(material, boundary_condition,
            formulation, mesh, function_spaces, solver, contact_formulation=contact_formulation)

        # BUILD THE TANGENT OPERATORS BEFORE HAND
        if material.mtype == "CoupleStressModel":
            I = np.eye(material.ndim,material.ndim)

            material.elasticity_tensor = Voigt(material.lamb*np.einsum('ij,kl',I,I)+material.mu*(np.einsum('ik,jl',I,I)+np.einsum('il,jk',I,I)),1)
            material.gradient_elasticity_tensor = 2.*material.eta*I
            material.coupling_tensor0 = np.zeros((material.elasticity_tensor.shape[0],
                material.gradient_elasticity_tensor.shape[0]))

            # print material.elasticity_tensor
            ngauss = function_spaces[0].AllGauss.shape[0]
            d0 = material.elasticity_tensor.shape[0]
            d1 = material.elasticity_tensor.shape[1]
            d2 = material.gradient_elasticity_tensor.shape[0]
            d3 = material.gradient_elasticity_tensor.shape[1]
            material.elasticity_tensors = np.tile(material.elasticity_tensor.ravel(),ngauss).reshape(ngauss,d0,d1)
            material.gradient_elasticity_tensors = np.tile(material.gradient_elasticity_tensor.ravel(),ngauss).reshape(ngauss,d2,d3)

        elif material.mtype == "IsotropicLinearFlexoelectricModel":
            I = np.eye(material.ndim,material.ndim)

            material.elasticity_tensor = Voigt(material.lamb*np.einsum('ij,kl',I,I)+material.mu*(np.einsum('ik,jl',I,I)+np.einsum('il,jk',I,I)),1)
            material.gradient_elasticity_tensor = 2.*material.eta*I
            material.coupling_tensor0 = np.zeros((material.elasticity_tensor.shape[0],
                material.gradient_elasticity_tensor.shape[0]))
            material.piezoelectric_tensor = material.P
            material.flexoelectric_tensor = material.f
            material.dielectric_tensor = -material.eps*np.eye(material.ndim,material.ndim)

            # print material.elasticity_tensor
            ngauss = function_spaces[0].AllGauss.shape[0]
            d0 = material.elasticity_tensor.shape[0]
            d1 = material.elasticity_tensor.shape[1]
            d2 = material.gradient_elasticity_tensor.shape[0]
            d3 = material.gradient_elasticity_tensor.shape[1]
            material.elasticity_tensors = np.tile(material.elasticity_tensor.ravel(),ngauss).reshape(ngauss,d0,d1)
            material.gradient_elasticity_tensors = np.tile(material.gradient_elasticity_tensor.ravel(),ngauss).reshape(ngauss,d2,d3)
            material.piezoelectric_tensors = np.tile(material.P.ravel(),ngauss).reshape(ngauss,material.P.shape[0],material.P.shape[1])
            material.flexoelectric_tensors = np.tile(material.f.ravel(),ngauss).reshape(ngauss,material.f.shape[0],material.f.shape[1])
            material.dielectric_tensors = np.tile(material.dielectric_tensor.ravel(),ngauss).reshape(ngauss,material.ndim,material.ndim)

            factor = -1.
            material.H_Voigt = np.zeros((ngauss,material.H_VoigtSize,material.H_VoigtSize))
            for i in range(ngauss):
                H1 = np.concatenate((material.elasticity_tensor,factor*material.piezoelectric_tensor),axis=1)
                H2 = np.concatenate((factor*material.piezoelectric_tensor.T,material.dielectric_tensor),axis=1)
                H_Voigt = np.concatenate((H1,H2),axis=0)
                material.H_Voigt[i,:,:] = H_Voigt


        return function_spaces, solver



    def Solve(self, formulation=None, mesh=None,
        material=None, boundary_condition=None,
        function_spaces=None, solver=None,
        contact_formulation=None):
        """Main solution routine for FEMSolver """

        # CHECK DATA CONSISTENCY
        mesh = formulation.meshes[0]
        #---------------------------------------------------------------------------#
        function_spaces, solver = self.__checkdata__(material, boundary_condition,
            formulation, mesh, function_spaces, solver, contact_formulation=contact_formulation)
        #---------------------------------------------------------------------------#
        caller = inspect.getouterframes(inspect.currentframe(), 2)[1][3]
        if caller != "Solve":
            self.PrintPreAnalysisInfo(mesh, formulation)
        #---------------------------------------------------------------------------#

        # INITIATE DATA FOR NON-LINEAR ANALYSIS
        NodalForces, Residual = np.zeros((mesh.points.shape[0]*formulation.nvar,1),dtype=np.float64), \
            np.zeros((mesh.points.shape[0]*formulation.nvar,1),dtype=np.float64)
        # SET NON-LINEAR PARAMETERS
        self.NRConvergence = { 'Increment_'+str(Increment) : [] for Increment in range(self.number_of_load_increments) }

        # ALLOCATE FOR SOLUTION FIELDS
        TotalDisp = np.zeros((mesh.points.shape[0],formulation.nvar,self.number_of_load_increments),dtype=np.float64)
        TotalW = np.zeros((formulation.meshes[1].points.shape[0],formulation.ndim,self.number_of_load_increments),dtype=np.float64)
        TotalS = np.zeros((formulation.meshes[2].points.shape[0],formulation.ndim,self.number_of_load_increments),dtype=np.float64)
        # TotalDisp = np.zeros((mesh.points.shape[0],int(formulation.ndim**2),self.number_of_load_increments),dtype=np.float64)

        # PRE-ASSEMBLY
        if caller != "Solve":
            print('Assembling the system and acquiring neccessary information for the analysis...')
        tAssembly=time()

        # APPLY DIRICHELT BOUNDARY CONDITIONS AND GET DIRICHLET RELATED FORCES
        boundary_condition.GetDirichletBoundaryConditions(formulation, mesh, material, solver, self)

        # GET BOUNDARY INFO
        self.GetBoundaryInfo(mesh,formulation,boundary_condition)

        # ALLOCATE FOR GEOMETRY - GetDirichletBoundaryConditions CHANGES THE MESH
        # SO EULERX SHOULD BE ALLOCATED AFTERWARDS
        Eulerx = np.copy(formulation.meshes[0].points)
        Eulerw = np.zeros_like(formulation.meshes[1].points)
        Eulers = np.zeros_like(formulation.meshes[1].points)
        Eulerp = np.zeros((formulation.meshes[0].points.shape[0]))

        # FIND PURE NEUMANN (EXTERNAL) NODAL FORCE VECTOR
        NeumannForces = boundary_condition.ComputeNeumannForces(mesh, material, function_spaces,
            compute_traction_forces=True, compute_body_forces=self.add_self_weight)
        # NeumannForces = np.zeros((mesh.points.shape[0]*formulation.ndim))

        # ASSEMBLE STIFFNESS MATRIX AND TRACTION FORCES
        if self.analysis_type != 'static':
            if formulation.fields == "couple_stress" or formulation.fields == "flexoelectric":
                K, TractionForces, _, M = formulation.Assemble(self, material, Eulerx, Eulerw, Eulers, Eulerp)
            elif formulation.fields == "mechanics" or formulation.fields == "electro_mechanics":
                # STANDARD MECHANINCS ELECTROMECHANICS DYNAMIC ANALYSIS ARE DISPATCHED HERE
                from Florence.FiniteElements.Assembly import Assemble
                fspace = function_spaces[0] if (mesh.element_type=="hex" or mesh.element_type=="quad") else function_spaces[1]
                K, TractionForces, _, M = Assemble(self, fspace, formulation, mesh, material,
                    Eulerx, Eulerp)
        else:
            K, TractionForces = formulation.Assemble(self, material, Eulerx, Eulerw, Eulers, Eulerp)[:2]

        print('Finished all pre-processing stage. Time elapsed was', time()-tAssembly, 'seconds')

        if self.analysis_type != 'static':
            boundary_condition.ConvertStaticsToDynamics(mesh, self.number_of_load_increments)
            TotalDisp, TotalW, TotalS = self.DynamicSolver(formulation, solver,
                K, M, NeumannForces, NodalForces, Residual,
                mesh, TotalDisp, TotalW, TotalS, Eulerx, Eulerw, Eulers, Eulerp, material, boundary_condition)
        else:
            TotalDisp, TotalW, TotalS = self.StaticSolver(formulation, solver,
                K, NeumannForces, NodalForces, Residual,
                mesh, TotalDisp, TotalW, TotalS, Eulerx, Eulerw, Eulers, Eulerp, material, boundary_condition)


        solution = self.__makeoutput__(mesh, TotalDisp, formulation, function_spaces, material)
        if formulation.fields == "couple_stress" or formulation.fields == "flexoelectric":
            solution.solW = TotalW
            solution.solS = TotalS
        return solution



    def StaticSolver(self, formulation, solver, K,
            NeumannForces, NodalForces, Residual,
            mesh, TotalDisp, TotalW, TotalS, Eulerx, Eulerw, Eulers, Eulerp, material, boundary_condition):


        LoadIncrement = self.number_of_load_increments
        LoadFactor = 1./LoadIncrement
        AppliedDirichletInc = np.zeros(boundary_condition.applied_dirichlet.shape[0],dtype=np.float64)
        Increment = 0

        post_process = PostProcess(formulation.ndim,formulation.nvar)
        post_process.SetAnalysis(analysis_type=self.analysis_type, analysis_nature=self.analysis_nature)

        # APPLY NEUMANN BOUNDARY CONDITIONS
        DeltaF = LoadFactor*NeumannForces
        Residual[:,:] += DeltaF

        K_b, F_b, _ = boundary_condition.ApplyDirichletGetReducedMatrices(K,Residual,
                boundary_condition.applied_dirichlet,LoadFactor=LoadFactor)


        for Increment in range(LoadIncrement):
            t_increment=time()
            # SOLVE THE SYSTEM
            sol = solver.Solve(K_b, F_b, reuse_factorisation=True)

            AppliedDirichletInc = LoadFactor*boundary_condition.applied_dirichlet
            dU = post_process.TotalComponentSol(sol, boundary_condition.columns_in,
                boundary_condition.columns_out, AppliedDirichletInc,0,K.shape[0])

            # STORE TOTAL SOLUTION DATA
            TotalDisp[:,:,Increment] += dU

            # REDUCED ACCUMULATED FORCE
            Residual[:,:] += DeltaF
            F_b = boundary_condition.ApplyDirichletGetReducedMatrices(K,Residual,
                AppliedDirichletInc,LoadFactor=1.0,
                only_residual=True)[boundary_condition.columns_in,0]
            # F_b = boundary_condition.GetReducedVectors(Residual,only_residual=True)[0]

            # UPDATE
            Eulerx += dU[:,:formulation.ndim]
            Eulerp += dU[:,-1]
            TotalW[:,:,Increment], TotalS[:,:,Increment] = formulation.GetAugmentedSolution(self,
                material, TotalDisp, Eulerx, Eulerw, Eulers, Eulerp)

            # LOG REQUESTS
            self.LogSave(formulation, TotalDisp, Increment)

            # BREAK AT A SPECIFICED LOAD INCREMENT IF ASKED FOR
            if self.break_at_increment != -1 and self.break_at_increment is not None:
                if self.break_at_increment == Increment:
                    if self.break_at_increment < LoadIncrement - 1:
                        print("\nStopping at increment {} as specified\n\n".format(Increment))
                        TotalDisp = TotalDisp[:,:,:Increment]
                        self.number_of_load_increments = Increment
                    break

            print('Finished Load increment', Increment, 'in', time()-t_increment, 'seconds\n')

        # ADD EACH INCREMENTAL CONTRIBUTION TO MAKE IT CONSISTENT WITH THE NONLINEAR ANALYSIS
        for i in range(TotalDisp.shape[2]-1,0,-1):
            TotalDisp[:,:,i] = np.sum(TotalDisp[:,:,:i+1],axis=2)

        # COMPUTE DISSIPATION OF ENERGY THROUGH TIME
        if self.compute_energy:
            energy_info = self.ComputeEnergy(formulation.function_spaces[0],mesh,material,formulation,Eulerx,Eulerp)
            formulation.strain_energy = energy_info[0]
            formulation.electrical_energy = energy_info[1]

        solver.CleanUp()

        return TotalDisp, TotalW, TotalS



    def DynamicSolver(self, formulation, solver, K, M,
            NeumannForces, NodalForces, Residual,
            mesh, TotalDisp, TotalW, TotalS, Eulerx, Eulerw, Eulers, Eulerp, material, boundary_condition):

        LoadIncrement = self.number_of_load_increments
        LoadFactor = self.total_time/LoadIncrement

        post_process = PostProcess(formulation.ndim,formulation.nvar)
        post_process.SetAnalysis(analysis_type=self.analysis_type, analysis_nature=self.analysis_nature)

        if NeumannForces.ndim == 2 and NeumannForces.shape[1]==1:
            tmp = np.zeros((NeumannForces.shape[0],LoadIncrement))
            tmp[:,0] = NeumannForces[:,0]
            NeumannForces = tmp

        dumU = np.zeros((mesh.points.shape[0]*formulation.nvar))
        dumU[boundary_condition.columns_out] = boundary_condition.applied_dirichlet[:,0]
        TotalDisp[:,:formulation.nvar,0] = dumU.reshape(mesh.points.shape[0],formulation.nvar)
        # INITIALISE VELOCITY AND ACCELERATION
        velocities     = np.zeros((mesh.points.shape[0]*formulation.ndim))
        accelerations  = np.zeros((mesh.points.shape[0]*formulation.ndim))
        # COMPUTE DAMPING MATRIX BASED ON MASS
        D = 0.0
        if self.include_physical_damping:
            D = self.damping_factor*M

        if formulation.fields == "electro_mechanics" or formulation.fields == "flexoelectric":
            # self.GetBoundaryInfo(mesh, formulation,boundary_condition)
            M_mech = M[self.mechanical_dofs,:][:,self.mechanical_dofs]
            if self.include_physical_damping:
                D_mech = D[self.mechanical_dofs,:][:,self.mechanical_dofs]
        else:
            M_mech = M
            D_mech = D

        # COMPUTE INITIAL ACCELERATION FOR TIME STEP 0
        Residual = np.zeros_like(Residual)
        InitResidual = Residual + NeumannForces[:,0][:,None]
        if formulation.fields == "electro_mechanics" or formulation.fields == "flexoelectric":
            accelerations[:] = solver.Solve(M_mech, -InitResidual[self.mechanical_dofs].ravel())
        else:
            accelerations[:] = solver.Solve(M, InitResidual.ravel() )

        # COMPUTE AUGMENTED K (INCLUDES INERTIA EFFECT)
        K          += (self.gamma/self.beta/LoadFactor)*D + (1./self.beta/LoadFactor**2)*M
        # GET REDUCED VARIABLES
        K_b, F_b, _ = boundary_condition.GetReducedMatrices(K,Residual)

        if self.lump_rhs:
            M_mech = M_mech.sum(axis=1).A.ravel() # FOR CSR
            # M_mech = M_mech.sum(axis=0).ravel() # FOR CSC
            if self.include_physical_damping:
                D_mech = D_mech.sum(axis=1).A.ravel()


        for Increment in range(1,LoadIncrement):
            t_increment=time()

            # FIXED INCREMENTAL DIRICHLET
            AppliedDirichletInc = boundary_condition.applied_dirichlet[:,Increment-1]

            # APPLY NEUMANN BOUNDARY CONDITIONS
            DeltaF = NeumannForces[:,Increment][:,None]
            NodalForces = DeltaF

            # ACCUMULATED FORCE
            if self.include_physical_damping:
                if self.lump_rhs:
                    Residual[self.mechanical_dofs,0] = (1./self.beta/LoadFactor**2)*M_mech*TotalDisp[:,:formulation.ndim,Increment-1].ravel() +\
                        (1./self.beta/LoadFactor)*M_mech*velocities + (0.5/self.beta - 1.)*M_mech*accelerations +\
                        (self.gamma/self.beta/LoadFactor)*D_mech*TotalDisp[:,:formulation.ndim,Increment-1].ravel() +\
                        (self.gamma/self.beta - 1.)*D_mech*velocities -\
                        LoadFactor*((1-self.gamma)-self.gamma*(0.5/self.beta - 1.))*D_mech*accelerations
                else:
                    Residual[self.mechanical_dofs,0] = (1./self.beta/LoadFactor**2)*M_mech.dot(TotalDisp[:,:formulation.ndim,Increment-1].ravel()) +\
                        (1./self.beta/LoadFactor)*M_mech.dot(velocities) + (0.5/self.beta - 1.)*M_mech.dot(accelerations) +\
                        (self.gamma/self.beta/LoadFactor)*D_mech.dot(TotalDisp[:,:formulation.ndim,Increment-1].ravel()) +\
                        (self.gamma/self.beta - 1.)*D_mech.dot(velocities) -\
                        LoadFactor*((1-self.gamma)-self.gamma*(0.5/self.beta - 1.))*D_mech.dot(accelerations)
            else:
                if self.lump_rhs:
                    Residual[self.mechanical_dofs,0] = (1./self.beta/LoadFactor**2)*M_mech*TotalDisp[:,:formulation.ndim,Increment-1].ravel() +\
                        (1./self.beta/LoadFactor)*M_mech*velocities + (0.5/self.beta - 1.)*M_mech*accelerations
                else:
                    Residual[self.mechanical_dofs,0] = (1./self.beta/LoadFactor**2)*M_mech.dot(TotalDisp[:,:formulation.ndim,Increment-1].ravel()) +\
                        (1./self.beta/LoadFactor)*M_mech.dot(velocities) + (0.5/self.beta - 1.)*M_mech.dot(accelerations)
            Residual += DeltaF


            # CHECK CONTACT AND ASSEMBLE IF DETECTED
            if self.has_contact:
                Eulerx = mesh.points + TotalDisp[:,:formulation.ndim,Increment-1]
                TractionForcesContact = np.zeros_like(Residual)
                TractionForcesContact = self.contact_formulation.AssembleTractions(mesh,material,Eulerx).ravel()*LoadFactor

                if formulation.fields == "electro_mechanics" or formulation.fields == "flexoelectric":
                    Residual[self.mechanical_dofs,0] -= TractionForcesContact
                elif formulation.fields == "mechanics" or formulation.fields == "couple_stress":
                    Residual[:,0] -= TractionForcesContact
                else:
                    raise NotImplementedError("Contact algorithm for {} is not available".format(formulation.fields))

            # REDUCED ACCUMULATED FORCE
            F_b = boundary_condition.ApplyDirichletGetReducedMatrices(K,Residual,
                boundary_condition.applied_dirichlet[:,Increment],LoadFactor=1.0,
                mass=M,only_residual=True)[boundary_condition.columns_in,0]

            # SOLVE THE SYSTEM
            sol = solver.Solve(K_b, F_b, reuse_factorisation=True)

            dU = post_process.TotalComponentSol(sol, boundary_condition.columns_in,
                boundary_condition.columns_out, AppliedDirichletInc,0,K.shape[0])

            # STORE TOTAL SOLUTION DATA
            TotalDisp[:,:,Increment] += dU

            # UPDATE VELOCITY AND ACCELERATION
            accelerations_old = np.copy(accelerations)
            accelerations = (1./self.beta/LoadFactor**2)*(TotalDisp[:,:formulation.ndim,Increment] -\
                TotalDisp[:,:formulation.ndim,Increment-1]).ravel() -\
                1./self.beta/LoadFactor*velocities + (1.-0.5/self.beta)*accelerations_old
            velocities += LoadFactor*(self.gamma*accelerations + (1-self.gamma)*accelerations_old)

            # UPDATE
            Eulerx += dU[:,:formulation.ndim]
            Eulerp += dU[:,-1]
            if formulation.fields == "couple_stress" or formulation.fields == "flexoelectric":
                TotalW[:,:,Increment], TotalW[:,:,Increment] = formulation.GetAugmentedSolution(self, material,
                    TotalDisp, Eulerx, Eulerw, Eulers, Eulerp)

            # LOG REQUESTS
            self.LogSave(formulation, TotalDisp, Increment)

            # BREAK AT A SPECIFICED LOAD INCREMENT IF ASKED FOR
            if self.break_at_increment != -1 and self.break_at_increment is not None:
                if self.break_at_increment == Increment:
                    if self.break_at_increment < LoadIncrement - 1:
                        print("\nStopping at increment {} as specified\n\n".format(Increment))
                        TotalDisp = TotalDisp[:,:,:Increment]
                        self.number_of_load_increments = Increment
                    break

            # STORE THE INFORMATION IF THE SOLVER BLOWS UP
            if Increment > 0:
                U0 = TotalDisp[:,:,Increment-1].ravel()
                U = TotalDisp[:,:,Increment].ravel()
                tol = 1e200 if Increment < 5 else 10.
                if np.isnan(norm(U)) or np.abs(U.max()/(U0.max()+1e-14)) > tol:
                    print("Solver blew up! Norm of incremental solution is too large")
                    TotalDisp = TotalDisp[:,:,:Increment]
                    self.number_of_load_increments = Increment
                    break

            print('Finished Load increment', Increment, 'in', time()-t_increment, 'seconds\n')

        # COMPUTE DISSIPATION OF ENERGY THROUGH TIME
        if self.compute_energy:
            energy_info = self.ComputeEnergy(formulation.function_spaces[0],mesh,material,formulation,Eulerx,Eulerp)
            formulation.strain_energy = energy_info[0]
            formulation.electrical_energy = energy_info[1]

        solver.CleanUp()

        return TotalDisp, TotalW, TotalS





    def ComputeEnergy(self,function_space,mesh,material,formulation,Eulerx,Eulerp):

        strain_energy = 0.
        electrical_energy = 0.
        for elem in range(mesh.nelem):
            LagrangeElemCoords = mesh.points[mesh.elements[elem,:],:]
            EulerElemCoords = Eulerx[mesh.elements[elem,:],:]
            ElectricPotentialElem = Eulerp[mesh.elements[elem,:]]

            energy = formulation.GetEnergy(function_space, material,
                LagrangeElemCoords, EulerElemCoords, ElectricPotentialElem, self, elem)
            strain_energy += energy[0]
            electrical_energy += energy[1]

        return strain_energy, electrical_energy



    def ComputeEnergyDissipation(self,function_space,mesh,material,formulation,fem_solver,
        Eulerx, TotalDisp, NeumannForces, M, velocities, Increment):

        internal_energy = 0.
        for elem in range(mesh.nelem):
            LagrangeElemCoords = mesh.points[mesh.elements[elem,:],:]
            EulerElemCoords = Eulerx[mesh.elements[elem,:],:]

            internal_energy += formulation.GetEnergy(function_space, material,
                LagrangeElemCoords, EulerElemCoords, fem_solver, elem)

        if formulation.fields == "flexoelectric":
            M_mech = M[self.mechanical_dofs,:][:,self.mechanical_dofs]
            kinetic_energy = 0.5*np.dot(velocities.ravel(),M_mech.dot(velocities.ravel()))
        else:
            kinetic_energy = 0.5*np.dot(velocities.ravel(),M.dot(velocities.ravel()))

        external_energy = np.dot(TotalDisp[:,:,Increment].ravel(),NeumannForces[:,Increment])

        total_energy = internal_energy + kinetic_energy - external_energy
        return total_energy, internal_energy, kinetic_energy, external_energy



    def ComputePowerDissipation(self,function_space,mesh,material,formulation,fem_solver, Eulerx, TotalDisp,
        NeumannForces, M, velocities, accelerations, Increment):

        internal_energy = 0.
        for elem in range(mesh.nelem):
            LagrangeElemCoords = mesh.points[mesh.elements[elem,:],:]
            EulerElemCoords    = Eulerx[mesh.elements[elem,:],:]
            VelocityElem       = velocities[mesh.elements[elem,:],:]

            internal_energy += formulation.GetLinearMomentum(function_space, material,
                LagrangeElemCoords, EulerElemCoords, VelocityElem, fem_solver, elem)

        if formulation.fields == "electro_mechanics":
            M_mech = M[self.mechanical_dofs,:][:,self.mechanical_dofs]
            kinetic_energy = np.dot(velocities.ravel(),M_mech.dot(accelerations.ravel()))
        else:
            kinetic_energy = np.dot(velocities.ravel(),M.dot(accelerations.ravel()))

        external_energy = np.dot(velocities.ravel(),NeumannForces[:,Increment])

        total_energy = internal_energy + kinetic_energy - external_energy
        return total_energy, internal_energy, kinetic_energy, external_energy



import numpy as np
from numpy import einsum
from Florence.Tensor import trace, Voigt
from .MaterialBase import Material
from Florence.LegendreTransform import LegendreTransform

class IsotropicElectroMechanics_201(Material):
    """Electromechanical model in terms of internal energy 
            W(C,D) = W_mn(C) + 1/2/eps_1 (FD0*FD0)
            W_mn(C) = u1*C:I+u2*G:I - 2*(u1+2*u2)*lnJ + lamb/2*(J-1)**2
    """
    
    def __init__(self, ndim, **kwargs):
        mtype = type(self).__name__
        super(IsotropicElectroMechanics_201, self).__init__(mtype, ndim, **kwargs)
        # REQUIRES SEPARATELY
        self.nvar = self.ndim+1
        self.energy_type = "internal_energy"
        self.legendre_transform = LegendreTransform()
        self.nature = "nonlinear"
        self.fields = "electro_mechanics"

        if self.ndim == 2:
            self.H_VoigtSize = 5
        elif self.ndim == 3:
            self.H_VoigtSize = 9

        # LOW LEVEL DISPATCHER
        self.has_low_level_dispatcher = False


    def Hessian(self,StrainTensors,ElectricDisplacementx,elem=0,gcounter=0):

        mu1 = self.mu1
        mu2 = self.mu2
        lamb = self.lamb
        eps_1 = self.eps_1

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]

        D  = ElectricDisplacementx.reshape(self.ndim,1)
        Dx = D.reshape(self.ndim)
        DD = np.dot(D.T,D)[0,0]

        self.elasticity_tensor = 2.*mu2/J*(2*einsum('ij,kl',b,b) - einsum('ik,jl',b,b) - einsum('il,jk',b,b)) +\
            2.*(mu1+2.*mu2)/J*( einsum("ik,jl",I,I)+einsum("il,jk",I,I) ) + \
            lamb*(2.*J-1.)*einsum("ij,kl",I,I) - lamb*(J-1)*( einsum("ik,jl",I,I)+einsum("il,jk",I,I) )

        self.coupling_tensor = J/eps_1*(einsum('ik,j',I,Dx) + einsum('i,jk',Dx,I))

        self.dielectric_tensor = J/eps_1*I 

        # TRANSFORM TENSORS TO THEIR ENTHALPY COUNTERPART
        E_Voigt, P_Voigt, C_Voigt = self.legendre_transform.InternalEnergyToEnthalpy(self.dielectric_tensor, 
            self.coupling_tensor, self.elasticity_tensor)

        # BUILD HESSIAN
        factor = -1.
        H1 = np.concatenate((C_Voigt,factor*P_Voigt),axis=1)
        H2 = np.concatenate((factor*P_Voigt.T,E_Voigt),axis=1)
        H_Voigt = np.concatenate((H1,H2),axis=0)

        return H_Voigt



    def CauchyStress(self,StrainTensors,ElectricDisplacementx,elem=0,gcounter=0):

        mu1 = self.mu1
        mu2 = self.mu2
        lamb = self.lamb
        eps_1 = self.eps_1

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]
        D = ElectricDisplacementx.reshape(self.ndim,1)

        if self.ndim==2:
            trb = trace(b) + 1
        else:
            trb = trace(b)

        simga_mech = 2.0*mu1/J*b + \
            2.0*mu2/J*(trb*b - np.dot(b,b)) -\
            2.0*(mu1+2*mu2)/J*I +\
            lamb*(J-1)*I
        sigma_electric = J/eps_1*np.dot(D,D.T)
        sigma = simga_mech + sigma_electric

        return sigma

    def ElectricDisplacementx(self,StrainTensors,ElectricFieldx,elem=0,gcounter=0):
        D = self.legendre_transform.GetElectricDisplacement(self, StrainTensors, ElectricFieldx, elem, gcounter)


        # SANITY CHECK FOR IMPLICIT COMPUTATUTAION OF D
        # I = StrainTensors['I']
        # J = StrainTensors['J'][gcounter]
        # E = ElectricFieldx.reshape(self.ndim,1)
        # eps_1 = self.eps_1
        # D_exact = eps_1/J*E
        # # print np.linalg.norm(D - D_exact)
        # return D_exact

        return D



from __future__ import division
import numpy as np
from numpy import einsum
from .MaterialBase import Material
from Florence.Tensor import trace, Voigt, makezero


class AnisotropicMooneyRivlin_1(Material):
    """A compressible transervely isotropic Mooney-Rivlin model with the energy given by:

            W(C) =  u1*(C:I) + u2*(G:I) - 2*(u1+2*u2+u3)*lnJ + lamb/2*(J-1)**2 + u3 FN.FN + u3 HN.HN

        where G=(H^T*H) and H=cofactor(F). The stress is zero at the origin

    """

    def __init__(self, ndim, **kwargs):
        mtype = type(self).__name__
        super(AnisotropicMooneyRivlin_1, self).__init__(mtype, ndim, **kwargs)
        self.nvar = self.ndim
        self.is_transversely_isotropic = True
        self.energy_type = "internal_energy"
        self.nature = "nonlinear"
        self.fields = "mechanics"
        
        if self.ndim==3:
            self.H_VoigtSize = 6
        else:
            self.H_VoigtSize = 3

        # LOW LEVEL DISPATCHER
        self.has_low_level_dispatcher = True
        # self.has_low_level_dispatcher = False

    def KineticMeasures(self,F,ElectricFieldx=0, elem=0):
        from Florence.MaterialLibrary.LLDispatch._AnisotropicMooneyRivlin_1_ import KineticMeasures
        return KineticMeasures(self, F, self.anisotropic_orientations[elem][:,None])



    def Hessian(self,StrainTensors,ElectricFieldx=0,elem=0,gcounter=0):

        mu1 = self.mu1
        mu2 = self.mu2
        mu3 = self.mu3
        lamb = self.lamb

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]
        F = StrainTensors['F'][gcounter]
        H = J*np.linalg.inv(F).T
        N = self.anisotropic_orientations[elem][:,None]
        HN = np.dot(H,N)[:,0]
        innerHN = einsum('i,i',HN,HN)
        outerHN = einsum('i,j',HN,HN)

        H_Voigt = 2.*mu2/J* ( 2.0*einsum('ij,kl',b,b) - einsum('ik,jl',b,b) - einsum('il,jk',b,b) ) + \
                2.*(mu1+2.*mu2+mu3)/J * ( einsum('ik,jl',I,I) + einsum('il,jk',I,I) ) + \
                lamb*(2.*J-1.)*einsum('ij,kl',I,I) - lamb*(J-1.) * ( einsum('ik,jl',I,I) + einsum('il,jk',I,I) ) - \
                4.*mu3/J * ( einsum('ij,kl',I,outerHN) + einsum('ij,kl',outerHN,I) ) + \
                2.*mu3/J*innerHN*(2.0*einsum('ij,kl',I,I) - einsum('ik,jl',I,I) - einsum('il,jk',I,I) ) + \
                2.*mu3/J * ( einsum('ik,jl',I,outerHN) + einsum('il,jk',I,outerHN) + \
                einsum('ik,jl',outerHN,I) + einsum('il,jk',outerHN,I) )  
                # 2.*mu3/J * ( einsum('il,j,k',I,HN,HN) + einsum('jl,i,k',I,HN,HN) + \
                # einsum('ik,j,l',I,HN,HN) + einsum('jk,i,l',I,HN,HN) ) 

        H_Voigt = Voigt(H_Voigt ,1)              
        
        return H_Voigt



    def CauchyStress(self,StrainTensors,ElectricFieldx=0,elem=0,gcounter=0):

        I = StrainTensors['I']
        J = StrainTensors['J'][gcounter]
        b = StrainTensors['b'][gcounter]
        F = StrainTensors['F'][gcounter]
        H = J*np.linalg.inv(F).T
        N = self.anisotropic_orientations[elem][:,None]
        FN = np.dot(F,N)[:,0]
        HN = np.dot(H,N)[:,0]
        outerFN = einsum('i,j',FN,FN)
        innerHN = einsum('i,i',HN,HN)
        outerHN = einsum('i,j',HN,HN)

        mu1 = self.mu1
        mu2 = self.mu2
        mu3 = self.mu3
        lamb = self.lamb

        if self.ndim == 3:
            trb = trace(b)
        elif self.ndim == 2:
            trb = trace(b) + 1

        stress = 2.*mu1/J*b + \
            2.*mu2/J*(trb*b - np.dot(b,b)) - \
            2.*(mu1+2.*mu2+mu3)/J*I + \
            lamb*(J-1)*I +\
            2.*mu3/J*outerFN +\
            2.*mu3/J*innerHN*I - 2.*mu3/J*outerHN

        return stress


from __future__ import print_function, division
import numpy as np
import scipy as sp
import warnings, sys

def GenerateCoordinates(boundary_elements,boundary_points,C,z):
    # Generate the entire coordinate system for higher order elements

    from Florence.FunctionSpace.OneDimensional.Line import LagrangeGaussLobatto, Lagrange
    # Get nodal position at the parent coordinate
    # eps = Lagrange(C,0)[2]
    eps = LagrangeGaussLobatto(C,0)[2]
    # Check if Gauss point and interior nodes coincide
    if eps[C]==z[int(z.shape[0]/2)]:
        sys.exit("Error: At least one Gauss point and an interior node coincide. Change the order of integration to one order higher or lower")

    # Allocate
    global_coord = np.zeros((boundary_points.shape[0]+C*boundary_elements.shape[0],2))
    for elem in range(0,len(boundary_elements)):
        # Get the two edge nodes of the element
        node_i = boundary_elements[elem,0]
        node_j = boundary_elements[elem,1]
        # Get their coordinates
        node_i_coord = boundary_points[node_i,0:2]
        node_j_coord = boundary_points[node_j,0:2]
        # Get length of the element
        le = sp.linalg.norm(node_j_coord-node_i_coord)
        angle = np.arccos((node_j_coord[0]-node_i_coord[0])/le)
        # for higher order elements divide the boundary element into corresponding pieces (i.e. interior nodes)
        eps_physical = le*(eps+1)/2.0  # eps_physical contains le, so don't include it

        coord = np.zeros((len(eps_physical),2))
        for i in range(0,len(eps_physical)):
            x_coord = node_i_coord[0]+eps_physical[i]*np.cos(angle)
            if node_j_coord[1] > node_i_coord[1]:
                y_coord = node_i_coord[1]+eps_physical[i]*np.sin(angle)
            else:
                y_coord = node_i_coord[1]-eps_physical[i]*np.sin(angle)

            coord[i,0] = x_coord
            coord[i,1] = y_coord

        global_coord[(C+1)*elem:(C+1)*elem+(C+1),0:] = coord[0:(C+1),0:]

    return global_coord


def CoordsJacobianRadiusatGaussPoints(boundary_elements,global_coord,C,Basis,dN,w):

    # Allocate
    XCO = np.zeros((w.shape[0],boundary_elements.shape[0])); YCO = np.zeros((w.shape[0],boundary_elements.shape[0]))
    nx = np.zeros((w.shape[0],boundary_elements.shape[0])); ny = np.zeros((w.shape[0],boundary_elements.shape[0]))
    Jacobian = np.zeros((w.shape[0],boundary_elements.shape[0]))

    # Loop over elements
    for elem in range(0,boundary_elements.shape[0]):
        # Loop over Gauss points
        for g in range(0,w.shape[0]):
            if elem != boundary_elements[-1,0]:
                XCO[g,elem] = np.dot(Basis[0:,g],global_coord[(C+1)*elem:(C+1)*(elem+1)+1,0])
                YCO[g,elem] = np.dot(Basis[0:,g],global_coord[(C+1)*elem:(C+1)*(elem+1)+1,1])
            elif elem == boundary_elements[-1,0]:
                XCO[g,elem] = np.dot(Basis[0:,g],np.append(global_coord[(C+1)*elem:(C+1)*(elem+1)+1,0],global_coord[0,0]))
                YCO[g,elem] = np.dot(Basis[0:,g],np.append(global_coord[(C+1)*elem:(C+1)*(elem+1)+1,1],global_coord[0,1]))

            # Get the derivatives at Gauss points
            if elem != boundary_elements[-1,0]:
                dx_by_dxi = np.dot(dN[0:,g],global_coord[(C+1)*elem:(C+1)*(elem+1)+1,0])
                dy_by_dxi = np.dot(dN[0:,g],global_coord[(C+1)*elem:(C+1)*(elem+1)+1,1])
            elif elem == boundary_elements[-1,0]:
                dx_by_dxi = np.dot(dN[0:,g],np.append(global_coord[(C+1)*elem:(C+1)*(elem+1)+1,0],global_coord[0,0]))
                dy_by_dxi = np.dot(dN[0:,g],np.append(global_coord[(C+1)*elem:(C+1)*(elem+1)+1,1],global_coord[0,1]))


            # dx_dy_by_dxi = np.dot(dN,coord) # vectorised version
            # Compute Jacobian
            Jacobian[g,elem] = np.sqrt(dx_by_dxi**2+dy_by_dxi**2)
            if Jacobian[g,elem]==0:
                warnings.warn("Jacobian is zero!")

            # Compute unit normal
            nx[g,elem] = 1.0/Jacobian[g,elem]*dy_by_dxi
            ny[g,elem] = -1.0/Jacobian[g,elem]*dx_by_dxi


    return Jacobian, nx, ny, XCO, YCO


def CoordsJacobianRadiusatGaussPoints_LM(boundary_elements,global_coord,C,Basis,dN,w,element_connectivity):

    # Allocate
    XCO = np.zeros((w.shape[0],boundary_elements.shape[0])); YCO = np.zeros((w.shape[0],boundary_elements.shape[0]))
    nx = np.zeros((w.shape[0],boundary_elements.shape[0])); ny = np.zeros((w.shape[0],boundary_elements.shape[0]))
    Jacobian = np.zeros((w.shape[0],boundary_elements.shape[0]))

    # Loop over elements
    for elem in range(0,boundary_elements.shape[0]):
        coord = global_coord[element_connectivity[elem]]
        coord = coord[:,0:-1]
        # Loop over Gauss points
        for g in range(0,w.shape[0]):
            if elem != boundary_elements[-1,0]:
                XCO[g,elem] = np.dot(Basis[0:,g],coord[:,0])
                YCO[g,elem] = np.dot(Basis[0:,g],coord[:,1])
            elif elem == boundary_elements[-1,0]:
                XCO[g,elem] = np.dot(Basis[0:,g],np.append(coord[:,0],coord[0,0]))
                YCO[g,elem] = np.dot(Basis[0:,g],np.append(coord[:,1],coord[0,1]))

            # Get the derivatives at Gauss points
            if elem != boundary_elements[-1,0]:
                dx_by_dxi = np.dot(dN[0:,g],coord[:,0])
                dy_by_dxi = np.dot(dN[0:,g],coord[:,1])
            elif elem == boundary_elements[-1,0]:
                dx_by_dxi = np.dot(dN[0:,g],np.append(coord[:,0],global_coord[0,0]))
                dy_by_dxi = np.dot(dN[0:,g],np.append(coord[:,1],global_coord[0,1]))


            # dx_dy_by_dxi = np.dot(dN,coord) # vectorised version
            # Compute Jacobian
            Jacobian[g,elem] = np.sqrt(dx_by_dxi**2+dy_by_dxi**2)
            if Jacobian[g,elem]==0:
                warnings.warn("Jacobian is zero!")

            # Compute unit normal
            nx[g,elem] = 1.0/Jacobian[g,elem]*dy_by_dxi
            ny[g,elem] = -1.0/Jacobian[g,elem]*dx_by_dxi



    return Jacobian, nx, ny, XCO, YCO


from .PreProcessBEM2D import GenerateCoordinates, CoordsJacobianRadiusatGaussPoints, CoordsJacobianRadiusatGaussPoints_LM
from .PostProcessBEM2D import InteriorPostProcess
from .Sort_BEM import Sort_BEM
from .WritePlotBEM2D import WritePlotBEM2D
from .GetBases import GetBases as GetBasesBEM2D
from .Assembly import AssemblyBEM2D


import numpy as np 
from Core.JacobiPolynomials.JacobiPolynomials import *


def hpBases(C,x,y):

	# Size of these bases is (p+1)*(p+2)/2. where p is the polynomial degree (here p is denoted by r)

	r = C+1
	i=r-2
	j=r-2

	nsize = int((r+1.)*(r+2.)/2.)
	Bases = np.zeros(nsize)

	l0 = 1.-x-y
	l1 = x
	l2 = y

	Bases[:3] = np.array([l0,l1,l2])
	# print Bases

	if C >0:
		# Edges
		e1 = l0*l1*JacobiPolynomials(i,l1-l0,1.,1.)[:,0]
		e2 = l1*l2*JacobiPolynomials(i,l2-l1,1.,1.)[:,0]
		e3 = l2*l0*JacobiPolynomials(i,l0-l2,1.,1.)[:,0]

		Bases[3:3+C] = e1; Bases[3+C:3+2*C] = e2; Bases[3+2*C:3+3*C] = e3
		# print Bases

	if C>1:
		# Interiors
		interior = []
		for i in range(0,r-2):
			for j in range(0,r-2):
				if i+j <= r-3:
					interior = np.append(interior,l0*l1*l2*JacobiPolynomials(i,2.*x/(1.-y)-1.,1.,1.)[-1]*(1-y)**i*JacobiPolynomials(j,2.*y-1.,2.*i+1,1)[-1])
					# print i, j

		Bases[3+3*C:] = interior

		# print interior
		# print Bases


	return Bases


def GradhpBases(C,x,y):

	# Size of these bases is (p+1)*(p+2)/2. where p is the polynomial degree (here p is denoted by r)

	r = C+1
	i=r-2
	j=r-2

	nsize = int((r+1.)*(r+2.)/2.)
	GradBases = np.zeros((nsize,2))

	l0 = 1.-x-y
	l1 = x
	l2 = y

	dl0dx = -1.
	dl1dx = 1.
	dl2dx = 0.

	dl0dy = -1.
	dl1dy = 0.
	dl2dy = 1.

	GradBases[:3,:] = np.array([
		[dl0dx,dl1dx,dl2dx],
		[dl0dy,dl1dy,dl2dy]
		]).T
	# print GradBases

	if C >0:
		# Edges
		# l1-l0 = 2.*x+y-1. 		# d(l1-l0)/dx = 2. and d(l1-l0)/dy = 1. 
		# l2-l1 = y-x 				# d(l2-l1)/dx = -1. and d(l2-l1)/dy = 1.
		# l0-l2 = 1.-x-2.*y 		# d(l0-l2)/dx = -1. and d(l0-l2)/dx = -2.

		dl1l0dx = 2.;	 dl1l0dy = 1. 
		dl2l1dx = -1.;	 dl2l1dy = 1.
		dl0l2dx = -1.;	 dl0l2dy = -2.

		de1dx = (-1.)*l1*JacobiPolynomials(i,l1-l0,1.,1.) + l0*(1.)*JacobiPolynomials(i,l1-l0,1.,1.) + l0*l1*DiffJacobiPolynomials(i,l1-l0,1.,1.,1)*dl1l0dx
		de2dx = (1.)*l2*JacobiPolynomials(i,l2-l1,1.,1.) + l1*l2*DiffJacobiPolynomials(i,l2-l1,1.,1.,1)*dl2l1dx
		de3dx = l2*(-1.)*JacobiPolynomials(i,l0-l2,1.,1.) + l2*l0*JacobiPolynomials(i,l0-l2,1.,1.)*dl0l2dx

		de1dy = (-1.)*l1*JacobiPolynomials(i,l1-l0,1.,1.) + l0*l1*DiffJacobiPolynomials(i,l1-l0,1.,1.,1)*dl1l0dy
		de2dy = l1*(1.)*JacobiPolynomials(i,l2-l1,1.,1.) + l1*l2*DiffJacobiPolynomials(i,l2-l1,1.,1.,1)*dl2l1dy
		de3dy = (1.)*l0*JacobiPolynomials(i,l0-l2,1.,1.) + l2*(-1.)*JacobiPolynomials(i,l0-l2,1.,1.) + l2*l0*DiffJacobiPolynomials(i,l0-l2,1.,1.,1)*dl2l1dy


		GradBases[3:3+C,0] = de1dx[:,0]; GradBases[3+C:3+2*C,0] = de2dx[:,0]; GradBases[3+2*C:3+3*C,0] = de3dx[:,0]
		GradBases[3:3+C,1] = de1dy[:,0]; GradBases[3+C:3+2*C,1] = de2dy[:,0]; GradBases[3+2*C:3+3*C,1] = de3dy[:,0]
		# print GradBases

	if C>1:
		# Interiors
		dinteriordx = []; dinteriordy = []
		for i in range(0,r-2):
			for j in range(0,r-2):
				if i+j <= r-3:
					didx = (-1.)*l1*l2*JacobiPolynomials(i,2.*x/(1.-y)-1.,1.,1.)[-1]*(1-y)**i*JacobiPolynomials(j,2.*y-1.,2.*i+1,1)[-1] +\
					l0*(1.)*l2*JacobiPolynomials(i,2.*x/(1.-y)-1.,1.,1.)[-1]*(1-y)**i*JacobiPolynomials(j,2.*y-1.,2.*i+1,1)[-1] +\
					l0*l1*l2*DiffJacobiPolynomials(i,2.*x/(1.-y)-1.,1.,1.,1)[-1]*(2./(1.-y))*(1-y)**i*JacobiPolynomials(j,2.*y-1.,2.*i+1,1)[-1]

					dinteriordx = np.append(dinteriordx,didx)

					didy = (-1.)*l1*l2*JacobiPolynomials(i,2.*x/(1.-y)-1.,1.,1.)[-1]*(1-y)**i*JacobiPolynomials(j,2.*y-1.,2.*i+1,1)[-1] +\
					l0*l1*(1.)*JacobiPolynomials(i,2.*x/(1.-y)-1.,1.,1.)[-1]*(1-y)**i*JacobiPolynomials(j,2.*y-1.,2.*i+1,1)[-1] +\
					l0*l1*l2*DiffJacobiPolynomials(i,2.*x/(1.-y)-1.,1.,1.,1)[-1]*(2.*x/(1.-y)**2)*(1-y)**i*JacobiPolynomials(j,2.*y-1.,2.*i+1,1)[-1] +\
					l0*l1*l2*JacobiPolynomials(i,2.*x/(1.-y)-1.,1.,1.)[-1]*i*(1-y)**(i-1)*(-1.)*JacobiPolynomials(j,2.*y-1.,2.*i+1,1)[-1] +\
					l0*l1*l2*JacobiPolynomials(i,2.*x/(1.-y)-1.,1.,1.)[-1]*(1-y)**i*DiffJacobiPolynomials(j,2.*y-1.,2.*i+1.,1.,1)[-1]*(2.)

					dinteriordy = np.append(dinteriordy,didy)

		GradBases[3+3*C:,0] = dinteriordx
		GradBases[3+3*C:,1] = dinteriordy
		# print GradBases


	return GradBases


from __future__ import print_function
from setuptools import setup
from setuptools import find_packages
import os, platform, sys, subprocess, imp
from distutils.command.clean import clean
from distutils.extension import Extension
from distutils.sysconfig import get_config_var, get_config_vars, get_python_inc, get_python_lib
import fnmatch
try:
    from Cython.Build import cythonize
except ImportError:
    raise ImportError("Could not import cython. Please install cython first")
try:
    import numpy as np
except ImportError:
    raise ImportError("Could not import numpy. Please install numpy first")

from multiprocessing import Pool, cpu_count



class FlorenceSetup(object):

    _os = sys.platform

    python_version = None
    python_interpreter = None
    python_implementation = None
    python_include_path = None
    python_ld_path = None

    numpy_version = None
    numpy_include_path = None

    fastor_include_path = None

    blas_version = "openblas"
    blas_include_path = None
    blas_ld_path = None

    fc_compiler = None
    cc_compiler = None
    cxx_compiler = None

    compiler_args = None
    fc_compiler_args = None
    cc_compiler_args = None
    cxx_compiler_args = None

    fc_version = None
    cc_version = None
    cxx_version = None

    extension_paths = None
    extension_postfix = None

    parallel = True
    num_cpu = cpu_count()

    kinematics_version = "-DUSE_AVX_VERSION"

    def __init__(self, _fc_compiler=None, _cc_compiler=None, _cxx_compiler=None,
        _blas=None, _kinematics_version=None, _fastor_path=None, _ncpu=1):

        # Get current working working directory
        self._pwd_ = os.path.dirname(os.path.realpath('__file__'))
        # Get python version and paths to header
        self.GetPythonPath()
        # Get numpy version and paths to header
        self.GetNumPyPath()
        # Get Fastor path
        self.GetFastorPath(_fastor_path)
        # Get BLAS version and paths
        self.GetBLAS(_blas)
        # Get kinematics version
        self.GetKinematicsVersion(_kinematics_version)
        # Set C/Fortran/C++ compiler
        self.SetCompiler(_fc_compiler,_cc_compiler,_cxx_compiler)
        # Set up compiler arguments for all extension modules
        self.SetCompilerArgs()
        # Collect all extension module paths
        self.CollectExtensionModulePaths()
        # Determine parallel builds
        self.SetParallelism(_ncpu)


    def GetPythonPath(self):

        if self._pwd_ is None:
            self._pwd_ = os.path.dirname(os.path.realpath('__file__'))

        # Get python version
        python_version = sys.version_info

        if python_version[:2] < (2, 7) or (3, 0) <= python_version[:2] < (3, 5):
            raise RuntimeError("Python version 2.7 or >= 3.5 required.")

        py_major = python_version.major
        py_minor = python_version.minor
        py_micro = python_version.micro
        self.python_interpreter  = 'python' + str(py_major) +'.' + str(py_minor)

        with_pymalloc = ""
        if get_config_var("ABIFLAGS") is not None:
            with_pymalloc = get_config_var("ABIFLAGS")
        self.python_interpreter += with_pymalloc

        # Get python implementation e.g. CPython, PyPy etc
        self.python_implementation = platform.python_implementation()

        # Get python include path
        self.python_include_path = get_python_inc()

        # Get python lib path
        # Note that we need the actual ld path where libpython.so/dylib/dll resides
        # and this seems like the only portable way at the moment
        lib_postfix = ".so"
        if "darwin" in self._os:
            lib_postfix = ".dylib"

        # The following search is split into two searches for /usr/lib and /usr/local
        # for speed purposes
        libpython = "lib"+self.python_interpreter+lib_postfix
        for root, _, filenames in os.walk('/usr/lib/'):
            for filename in fnmatch.filter(filenames, libpython):
                self.python_ld_path = os.path.join(root, filename).rsplit(libpython)[0]
                break
        if self.python_ld_path is None:
            for root, _, filenames in os.walk('/usr/local/'):
                for filename in fnmatch.filter(filenames, libpython):
                    self.python_ld_path = os.path.join(root, filename).rsplit(libpython)[0]
                    break

        # For conda envs change python_ld_path
        if "conda" in sys.version or "Continuum" in sys.version or "Intel" in sys.version:
            self.python_ld_path = get_config_var("LIBDIR")

        # Sanity check
        if self.python_ld_path is None:
            try:
                self.python_ld_path = get_config_var("LIBDIR")
            except:
                raise RuntimeError("Could not find libpython")


        # Get postfix for extensions
        self.extension_postfix = get_config_vars()['SO'][1:]
        if self.extension_postfix is None:
            if "darwin" in self._os:
                self.extension_postfix = "dylib"
            else:
                self.extension_postfix = "so"


    def GetNumPyPath(self):
        numpy_version = np.__version__
        if int(numpy_version.split('.')[0])==1 and int(numpy_version.split('.')[1]) < 8:
            raise RuntimeError("Numpy version >= 1.8 required")
        elif int(numpy_version.split('.')[0]) < 1:
            raise RuntimeError("Numpy version >= 1.8 required")
        self.numpy_include_path = np.get_include()


    def GetFastorPath(self, _fastor_path=None):
        if _fastor_path is not None:
            if os.path.isdir(_fastor_path):
                self.fastor_include_path = _fastor_path
                return

        if os.path.isdir("/usr/local/include/Fastor"):
            self.fastor_include_path = "/usr/local/include/Fastor"
        else:
            raise RuntimeError("Florence expects Fastor headers to be under '/usr/local/include/Fastor/' directory")
            return


    def GetBLAS(self, _blas=None):

        if _blas is not None:
            self.blas_version = _blas

        lib_postfix = ".so"
        if "darwin" in self._os:
            lib_postfix = ".dylib"

        if self.blas_version == "openblas":

            dirs = ["/opt/OpenBLAS/lib","/usr/local/Cellar/openblas/","/usr/lib/","/usr/local/lib/"]
            aux_path = "/usr/local/Cellar/openblas"
            if os.path.isdir(aux_path):
                files = os.listdir(aux_path)
                if self.blas_version+lib_postfix not in files:
                    for d in files:
                        if os.path.isdir(os.path.join(aux_path,d)):
                            files2 = os.listdir(os.path.join(aux_path,d))
                            if self.blas_version+lib_postfix not in files:
                                if os.path.isdir(os.path.join(aux_path,d,"lib")):
                                    dirs.append(os.path.join(aux_path,d,"lib"))

            found_blas = False
            for d in dirs:
                if os.path.isdir(d):
                    libs = os.listdir(d)
                    for blas in libs:
                        if self.blas_version+lib_postfix in blas:
                            self.blas_ld_path = d
                            self.blas_include_path = d.replace("lib","include")
                            found_blas = True
                            break
                    if found_blas:
                        break

        elif self.blas_version == "mkl":
            self.blas_ld_path = get_config_vars()['LIBDIR']
            self.blas_include_path = get_config_vars()['INCLUDEDIR']
            self.blas_version = "mkl_rt"


        # Fall back to reference BLAS if OpenBLAS is not found
        if self.blas_ld_path == None:
            self.blas_version = "blas"
            self.blas_ld_path = "/usr/lib"
            self.blas_include_path = "/usr/include/"


    def GetKinematicsVersion(self,_kinematics_version=None):
        if _kinematics_version == "scalar" or _kinematics_version == "None":
            self.kinematics_version = ""
        else:
            self.kinematics_version = "-DUSE_AVX_VERSION"


    def SetCompiler(self, _fc_compiler=None, _cc_compiler=None, _cxx_compiler=None):

        if not "darwin" in self._os and not "linux" in self._os and not "cygwin" in self._os:
            raise RuntimeError("Florence is not yet tested on any other platform apart from Linux & macOS")

        self.fc_compiler = _fc_compiler
        self.cc_compiler = _cc_compiler
        self.cxx_compiler = _cxx_compiler

        if self.fc_compiler is None:
            self.fc_compiler = "gfortran"

        if self.cc_compiler is None:
            self.cc_compiler = get_config_vars()['CC'].split(' ')[0]

        if self.cxx_compiler is None:
            self.cxx_compiler = get_config_vars()['CXX'].split(' ')[0]

        # Sanity check
        if self.cc_compiler is None:
            self.cc_compiler = "gcc"
        if self.cxx_compiler is None:
            self.cxx_compiler = "g++"


    def SetCompilerArgs(self):

        if "darwin" in self._os:
            self.cxx_version = "-std=c++14"
        else:
            self.cxx_version = "-std=c++11"

        # Generic compiler arguments
        self.compiler_args = "PYTHON_VERSION=" + self.python_interpreter + " PYTHON_INCLUDE_PATH=" + \
            self.python_include_path + " PYTHON_LD_PATH=" + self.python_ld_path + \
            " NUMPY_INCLUDE_PATH=" + self.numpy_include_path + \
            " FASTOR_INCLUDE_PATH=" + self.fastor_include_path +\
            " BLAS_VERSION=" + self.blas_version + " BLAS_INCLUDE_PATH="+ self.blas_include_path + \
            " BLAS_LD_PATH=" + self.blas_ld_path + " EXT_POSTFIX=" + self.extension_postfix +\
            " CXXSTD=" + self.cxx_version + " KINEMATICS=" + self.kinematics_version

        self.compiler_args = self.compiler_args + ' REMOVE="rm -rf" MOVE=mv '

        self.fc_compiler_args = "FC=" + self.fc_compiler + " " + self.compiler_args
        self.cc_compiler_args = "CC=" + self.cc_compiler + " " + self.compiler_args
        self.cxx_compiler_args = "CXX=" + self.cxx_compiler + " " + self.compiler_args

        self.compiler_args = "FC=" + self.fc_compiler + " " + "CC=" + self.cc_compiler + " " +\
            "CXX=" + self.cxx_compiler + " " + self.compiler_args



    def CollectExtensionModulePaths(self):
        # All modules paths should be specified in the following list
        _pwd_ = os.path.join(self._pwd_,"Florence")

        tensor_path = os.path.join(_pwd_,"Tensor")
        jacobi_path = os.path.join(_pwd_,"FunctionSpace","JacobiPolynomials")
        bp_path = os.path.join(_pwd_,"FunctionSpace","OneDimensional","_OneD")
        km_path = os.path.join(_pwd_,"FiniteElements","LocalAssembly","_KinematicMeasures_")
        gm_path = os.path.join(_pwd_,"VariationalPrinciple","_GeometricStiffness_")
        cm_path = os.path.join(_pwd_,"VariationalPrinciple","_ConstitutiveStiffness_")
        tm_path = os.path.join(_pwd_,"VariationalPrinciple","_Traction_")
        mm_path = os.path.join(_pwd_,"VariationalPrinciple","_Mass_")
        material_path = os.path.join(_pwd_,"MaterialLibrary","LLDispatch")
        assemble_path = os.path.join(_pwd_,"FiniteElements","Assembly","_Assembly_")

        self.extension_paths = [tensor_path,jacobi_path,bp_path,
            km_path,gm_path,cm_path,tm_path,mm_path,material_path,assemble_path]
        # self.extension_paths = [tensor_path]
        # self.extension_paths = [assemble_path]

    def SetParallelism(self,_ncpu=1):
        if _ncpu > 1:
            self.parallel = True
            self.num_cpu = int(_ncpu)
        elif _ncpu == 1:
            self.parallel = False
            self.num_cpu = 1

    def SourceClean(self):

        assert self.extension_paths != None

        source_clean_cmd = 'make source_clean ' + self.compiler_args
        for _path in self.extension_paths:
            if "_Assembly_" not in _path and "LLDispatch" not in _path:
                execute('cd '+_path+' && '+source_clean_cmd)
            elif "LLDispatch" in _path:
                execute('cd '+_path+' && echo rm -rf *.cpp CythonSource/*.cpp && rm -rf *.cpp CythonSource/*.cpp')
            elif "Assembly" in _path:
                execute('cd '+_path+' && echo rm -rf *.cpp && rm -rf *.cpp')

    def Clean(self):

        assert self.extension_paths != None

        self.SourceClean()
        # You need to run both make clean and rm -rf as some modules relocate the shared libraries
        clean_cmd = 'make clean ' + self.compiler_args
        for _path in self.extension_paths:
            if "LLDispatch" not in _path:
                execute('cd '+_path+' && echo rm -rf *.'+self.extension_postfix+' && '+
                    clean_cmd+' && rm -rf *.'+self.extension_postfix)
            elif "LLDispatch" in _path:
                execute('cd '+_path+
                    ' && echo rm -rf *.'+self.extension_postfix+' CythonSource/*.'+self.extension_postfix
                    +' && rm -rf *.'+self.extension_postfix+' CythonSource/*.'+
                    self.extension_postfix)
            else:
                execute('cd '+_path+' && echo rm -rf *.'+self.extension_postfix+' && rm -rf *.'+self.extension_postfix)

        # # clean all crude and ext libs if any - this is dangerous if setup.py is ever invoked from outside the directory
        # # which is certainly never the case
        # execute("find . -name \*.so -delete")
        # execute("find . -name \*.pyc -delete")


    def Build(self):


        low_level_material_list = [ "_LinearElastic_",
                                    "_NeoHookean_",
                                    "_MooneyRivlin_",
                                    "_NearlyIncompressibleMooneyRivlin_",
                                    "_AnisotropicMooneyRivlin_1_",
                                    "_ExplicitMooneyRivlin_",
                                    "_IsotropicElectroMechanics_0_",
                                    "_IsotropicElectroMechanics_3_",
                                    "_SteinmannModel_",
                                    "_IsotropicElectroMechanics_101_",
                                    "_IsotropicElectroMechanics_105_",
                                    "_IsotropicElectroMechanics_106_",
                                    "_IsotropicElectroMechanics_107_",
                                    "_IsotropicElectroMechanics_108_",
                                    "_IsotropicElectroMechanics_109_",
                                    "_Piezoelectric_100_",
                                    "_ExplicitIsotropicElectroMechanics_108_",
                                ]

        # low_level_material_list = ["_IsotropicElectroMechanics_109_"]
        # low_level_material_list = ["_NeoHookean_"]
        # low_level_material_list = ["_LinearElastic_"]
        # low_level_material_list = ["_ExplicitMooneyRivlin_"]


        assert self.extension_paths != None

        if self.parallel == False:
            for _path in self.extension_paths:
                if "LLDispatch" not in _path and "_Assembly_" not in _path:
                    execute('cd '+_path+' && make ' + self.compiler_args)
                if "LLDispatch" in _path:
                    _cmds = []
                    for material in low_level_material_list:
                        material = material.lstrip('_').rstrip('_')
                        execute('cd '+_path+' && make ' + self.compiler_args + " MATERIAL=" + material)

                elif "_Assembly_" in _path:

                    # Sparse and RHS assembler
                    execute('cd '+_path+' && make cython_assembler_build ' + self.compiler_args)

                    ll_material_mech = low_level_material_list[:6]
                    ll_material_electro_mech = low_level_material_list[6:]

                    ll_material_mech.remove("_ExplicitMooneyRivlin_")
                    ll_material_electro_mech.remove("_IsotropicElectroMechanics_109_")
                    ll_material_electro_mech.remove("_ExplicitIsotropicElectroMechanics_108_")

                    # ll_material_mech = []
                    # ll_material_electro_mech = low_level_material_list
                    # ll_material_mech = low_level_material_list
                    # ll_material_electro_mech = []

                    execute('cd '+_path+' && python AOT_Assembler.py clean')
                    execute('cd '+_path+' && python AOT_Assembler.py configure')

                    for material in ll_material_mech:
                        execute('cd '+_path+' && make ' + self.compiler_args + " ASSEMBLY_NAME=_LowLevelAssemblyDF_"  + material)
                    for material in ll_material_electro_mech:
                        execute('cd '+_path+' && make ' + self.compiler_args + " ASSEMBLY_NAME=_LowLevelAssemblyDPF_" + material)

                    execute('cd '+_path+' && python AOT_Assembler.py clean')

                    # Explicit assembler
                    execute('cd '+_path+' && make ' + self.compiler_args +\
                        " ASSEMBLY_NAME=_LowLevelAssemblyExplicit_DF_DPF_ CONDF_INC=../../../VariationalPrinciple/_Traction_/\
                        CONDF_INC=../../../VariationalPrinciple/_Traction_/")

                    # Perfect Laplacian assembler
                    execute('cd '+_path+' && make ' + self.compiler_args +\
                        " ASSEMBLY_NAME=_LowLevelAssemblyPerfectLaplacian_ ")

        else:
            # Parallel build
            pool = Pool(self.num_cpu)

            # All modules apart from LLDispatch and _Assembly_ modules
            _cmds = []
            for _path in self.extension_paths:
                if "LLDispatch" not in _path and "_Assembly_" not in _path:
                    _cmds.append('cd '+_path+' && make ' + self.compiler_args)

            pool.map(execute,_cmds)


            # LLDispatch modules
            _cmds = []
            for _path in self.extension_paths:
                if "LLDispatch" in _path:
                    for material in low_level_material_list:
                        material = material.lstrip('_').rstrip('_')
                        _cmds.append('cd '+_path+' && make ' + self.compiler_args + " MATERIAL=" + material)

            pool.map(execute,_cmds)


            # _Assembly_ modules
            _cmds = []
            for _path in self.extension_paths:
                if "_Assembly_" in _path:

                    ll_material_mech = low_level_material_list[:6]
                    ll_material_electro_mech = low_level_material_list[6:]

                    ll_material_mech.remove("_ExplicitMooneyRivlin_")
                    ll_material_electro_mech.remove("_IsotropicElectroMechanics_109_")
                    ll_material_electro_mech.remove("_ExplicitIsotropicElectroMechanics_108_")

                    # ll_material_mech = []
                    # ll_material_electro_mech = low_level_material_list
                    # ll_material_mech = low_level_material_list
                    # ll_material_electro_mech = []

                    execute('cd '+_path+' && python AOT_Assembler.py clean')
                    execute('cd '+_path+' && python AOT_Assembler.py configure')

                    for material in ll_material_mech:
                        _cmds.append('cd '+_path+' && make ' + self.compiler_args + " ASSEMBLY_NAME=_LowLevelAssemblyDF_"  + material)
                    for material in ll_material_electro_mech:
                        _cmds.append('cd '+_path+' && make ' + self.compiler_args + " ASSEMBLY_NAME=_LowLevelAssemblyDPF_" + material)

            pool.map(execute,_cmds)
            execute('cd '+_path+' && python AOT_Assembler.py clean')


            # Modules built sequentially
            for _path in self.extension_paths:
                if "_Assembly_" in _path:
                    # Sparse and RHS assembler
                    execute('cd '+_path+' && make cython_assembler_build ' + self.compiler_args)

                    # Explicit assembler
                    execute('cd '+_path+' && make ' + self.compiler_args +\
                        " ASSEMBLY_NAME=_LowLevelAssemblyExplicit_DF_DPF_ CONDF_INC=../../../VariationalPrinciple/_Traction_/\
                        CONDF_INC=../../../VariationalPrinciple/_Traction_/")

                    # Perfect Laplacian assembler
                    execute('cd '+_path+' && make ' + self.compiler_args +\
                        " ASSEMBLY_NAME=_LowLevelAssemblyPerfectLaplacian_ ")

            pool.close()


        # Get rid of cython sources
        sys.stdout = open(os.devnull, 'w')
        self.SourceClean()
        sys.stdout = sys.__stdout__


        sys.path.insert(1,self._pwd_)
        from Florence import Mesh, MaterialLibrary, FEMSolver
        from Florence.VariationalPrinciple import VariationalPrinciple


    def Install(self):
        var = raw_input("This includes florence in your python path. Do you agree (y/n): ")
        if var=="n" or "no" in var:
            return
        execute('export PYTHONPATH="$HOME/florence:$PYTHONPATH" >> ~/.profile && source ~/.profile')
        execute('export PYTHONPATH="' + self._pwd_ + ':$PYTHONPATH" >> ~/.profile && source ~/.profile')



def execute(_cmd):
    _process = subprocess.Popen(_cmd, shell=True)
    _process.wait()



def setup_package():

    _fc_compiler = None
    _cc_compiler = None
    _cxx_compiler = None
    _blas = None
    _kinematics_version = None
    _fastor_path = None
    _ncpu = cpu_count()

    args = sys.argv

    # should be either source_clean, clean or build
    _op = None

    if len(args) > 1:
        for arg in args:
            if arg == "source_clean" or arg == "clean" or arg == "build" or arg=="manual_install":
                if _op is not None:
                    raise RuntimeError("Multiple conflicting arguments passed to setup")
                _op = arg

            if "FC" in arg:
                _fc_compiler = arg.split("=")[-1]
            elif "CC" in arg:
                _cc_compiler = arg.split("=")[-1]
            elif "CXX" in arg:
                _cxx_compiler = arg.split("=")[-1]
            if "BLAS" in arg:
                _blas = arg.split("=")[-1]
            if "KINEMATICS" in arg:
                _kinematics_version = arg.split("=")[-1]
            if "FASTORPATH=" in arg:
                _fastor_path = int(arg.split("=")[-1])
            if "np=" in arg:
                _ncpu = int(arg.split("=")[-1])

    setup_instance = FlorenceSetup(_fc_compiler=_fc_compiler,
        _cc_compiler=_cc_compiler, _cxx_compiler=_cxx_compiler, _blas=_blas,
        _kinematics_version=_kinematics_version, _fastor_path=_fastor_path,_ncpu=_ncpu)

    if _op == "source_clean":
        setup_instance.SourceClean()
    elif _op == "clean":
        setup_instance.Clean()
    elif _op == "manual_install":
        setup_instance.Install()
    else:
        # setup_instance.Build()

        with open("README.md", "r") as fh:
            long_description = fh.read()

        setup(
                ext_modules = setup_instance.Build(),
                name = "Florence",
                version = "0.1",
                description = """A Python based computational framework for integrated computer aided design,
                    curvilinear mesh generation and finite and boundary element methods for linear and nonlinear
                    analysis of solids and coupled multiphysics problems""",
                long_description=long_description,
                author="Roman Poya",
                author_email = "roman_poya@yahoo.com",
                url = "https://github.com/romeric/florence",
                license="MIT",
                platforms=["Linux", "macOS", "Unix", "Windows"],
                python_requires='>=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*',
                install_requires=[
                  'cython>=0.23',
                  'numpy>=1.9',
                  'scipy>=0.14',
                  'scikit-umfpack>=0.2'],
                packages=find_packages(),
                include_package_data=True,
                package_data={'': ['*.pyx', '*.pxd', '*.h', '*.hpp', '*.py', '*.md', 'Makefile']},
            )




if __name__ == "__main__":
    setup_package()



import numpy as np
from copy import deepcopy
from warnings import warn
import gc
from time import time
import multiprocessing as MP

from Florence.Tensor import itemfreq, makezero, unique2d, remove_duplicates_2D
import Florence.ParallelProcessing.parmap as parmap
from .GetInteriorCoordinates import GetInteriorNodesCoordinates

#--------------------------------------------------------------------------------------------------------------------------#
# SUPPLEMENTARY FUNCTIONS
def ElementLoopHex(elem,elements,points,MeshType,eps,Neval):
    xycoord_higher = GetInteriorNodesCoordinates(points[elements[elem,:],:],MeshType,elem,eps,Neval)
    return xycoord_higher


def HighOrderMeshHex(C, mesh, Decimals=10, equally_spaced=False, check_duplicates=True,
    Zerofy=True, Parallel=False, nCPU=1, ComputeAll=True):

    from Florence.FunctionSpace import Hex, HexES
    from Florence.QuadratureRules import GaussLobattoPointsHex
    from Florence.QuadratureRules.EquallySpacedPoints import EquallySpacedPoints
    from Florence.MeshGeneration.NodeArrangement import NodeArrangementHex


    # SWITCH OFF MULTI-PROCESSING FOR SMALLER PROBLEMS WITHOUT GIVING A MESSAGE
    if (mesh.elements.shape[0] < 500) and (C < 5):
        Parallel = False
        nCPU = 1

    if not equally_spaced:
        eps = GaussLobattoPointsHex(C)
        # COMPUTE BASES FUNCTIONS AT ALL NODAL POINTS
        Neval = np.zeros((8,eps.shape[0]),dtype=np.float64)
        hpBases = Hex.LagrangeGaussLobatto
        for i in range(8,eps.shape[0]):
            Neval[:,i] = hpBases(0,eps[i,0],eps[i,1],eps[i,2])[:,0]
    else:
        eps = EquallySpacedPoints(4,C)
        # COMPUTE BASES FUNCTIONS AT ALL NODAL POINTS
        Neval = np.zeros((8,eps.shape[0]),dtype=np.float64)
        hpBases = HexES.Lagrange
        for i in range(8,eps.shape[0]):
            Neval[:,i] = hpBases(0,eps[i,0],eps[i,1],eps[i,2])[:,0]

    # THIS IS NECESSARY FOR REMOVING DUPLICATES
    makezero(Neval, tol=1e-12)

    nodeperelem = mesh.elements.shape[1]
    renodeperelem = int((C+2)**3)
    left_over_nodes = renodeperelem - nodeperelem

    reelements = -1*np.ones((mesh.elements.shape[0],renodeperelem),dtype=np.int64)
    reelements[:,:8] = mesh.elements
    # TOTAL NUMBER OF (INTERIOR+EDGE+FACE) NODES
    iesize = renodeperelem - 8
    repoints = np.zeros((mesh.points.shape[0]+iesize*mesh.elements.shape[0],3),dtype=np.float64)
    repoints[:mesh.points.shape[0],:]=mesh.points


    telements = time()

    xycoord_higher=[]; ParallelTuple1=[]
    if Parallel:
        # GET HIGHER ORDER COORDINATES - PARALLEL
        ParallelTuple1 = parmap.map(ElementLoopTet,np.arange(0,mesh.elements.shape[0]),mesh.elements,mesh.points,'hex',eps,
            Neval,pool=MP.Pool(processes=nCPU))

    maxNode = np.max(reelements)
    for elem in range(0,mesh.elements.shape[0]):
        # maxNode = np.max(reelements) # BIG BOTTLENECK
        if Parallel:
            xycoord_higher = ParallelTuple1[elem]
        else:
            xycoord =  mesh.points[mesh.elements[elem,:],:]
            # GET HIGHER ORDER COORDINATES
            xycoord_higher = GetInteriorNodesCoordinates(xycoord,'hex',elem,eps,Neval)

        # EXPAND THE ELEMENT CONNECTIVITY
        newElements = np.arange(maxNode+1,maxNode+1+left_over_nodes)
        reelements[elem,8:] = newElements
        # INSTEAD COMPUTE maxNode BY INDEXING
        maxNode = newElements[-1]

        repoints[mesh.points.shape[0]+elem*iesize:mesh.points.shape[0]+(elem+1)*iesize] = xycoord_higher[8:,:]

    if Parallel:
        del ParallelTuple1

    telements = time()-telements
    #--------------------------------------------------------------------------------------
    # NOW REMOVE DUPLICATED POINTS
    tnodes = time()

    nnode_linear = mesh.points.shape[0]
    # KEEP ZEROFY ON, OTHERWISE YOU GET STRANGE BEHVAIOUR
    # rounded_repoints = makezero(repoints[nnode_linear:,:].copy())
    rounded_repoints = repoints[nnode_linear:,:].copy()
    makezero(rounded_repoints)
    rounded_repoints = np.round(rounded_repoints,decimals=Decimals)

    _, idx_repoints, inv_repoints = unique2d(rounded_repoints,order=False,
        consider_sort=False,return_index=True,return_inverse=True)
    # idx_repoints.sort()
    del rounded_repoints#, flattened_repoints

    idx_repoints = np.concatenate((np.arange(nnode_linear),idx_repoints+nnode_linear))
    repoints = repoints[idx_repoints,:]

    unique_reelements, inv_reelements = np.unique(reelements[:,8:],return_inverse=True)
    unique_reelements = unique_reelements[inv_repoints]
    reelements = unique_reelements[inv_reelements]
    reelements = reelements.reshape(mesh.elements.shape[0],renodeperelem-8)
    reelements = np.concatenate((mesh.elements,reelements),axis=1)

    # SANITY CHECK fOR DUPLICATES
    #---------------------------------------------------------------------#
    # NOTE THAT THIS REMAPS THE ELEMENT CONNECTIVITY FOR THE WHOLE MESH
    # AND AS A RESULT THE FIRST FEW COLUMNS WOULD NO LONGER CORRESPOND TO
    # LINEAR CONNECTIVITY
    if check_duplicates:
        last_shape = repoints.shape[0]
        deci = int(Decimals)-2
        if Decimals < 6:
            deci = Decimals
        repoints, idx_repoints, inv_repoints = remove_duplicates_2D(repoints, decimals=deci)
        unique_reelements, inv_reelements = np.unique(reelements,return_inverse=True)
        unique_reelements = unique_reelements[inv_repoints]
        reelements = unique_reelements[inv_reelements]
        reelements = reelements.reshape(mesh.elements.shape[0],renodeperelem)
        if last_shape != repoints.shape[0]:
            warn('Duplicated points generated in high order mesh. Lower the "Decimals". I have fixed it for now')
    #---------------------------------------------------------------------#

    tnodes = time() - tnodes
    #------------------------------------------------------------------------------------------


    # USE ALTERNATIVE APPROACH TO GET MESH EDGES AND FACES
    reedges = np.zeros((mesh.edges.shape[0],C+2))
    fsize = int((C+2.)*(C+3.)/2.)
    refaces = np.zeros((mesh.faces.shape[0],fsize),dtype=mesh.faces.dtype)


    if ComputeAll == True:
        #------------------------------------------------------------------------------------------
        # BUILD FACES NOW
        tfaces = time()

        refaces = np.zeros((mesh.faces.shape[0],fsize))
        # DO NOT CHANGE THE FACES, BY RECOMPUTING THEM, AS THE LINEAR FACES CAN COME FROM
        # AN EXTERNAL MESH GENERATOR, WHOSE ORDERING MAY NOT BE THE SAME, SO JUST FIND WHICH
        # ELEMENTS CONTAIN THESE FACES
        face_to_elements = mesh.GetElementsWithBoundaryFacesHex()
        node_arranger = NodeArrangementHex(C)[0]

        refaces = reelements[face_to_elements[:,0][:,None],node_arranger[face_to_elements[:,1],:]].astype(mesh.faces.dtype)

        tfaces = time()-tfaces
        #------------------------------------------------------------------------------------------

        #------------------------------------------------------------------------------------------
        # BUILD EDGES NOW
        tedges = time()

        # BUILD A 2D MESH
        from Florence import Mesh
        tmesh = Mesh()
        tmesh.element_type = "quad"
        tmesh.elements = refaces
        tmesh.nelem = tmesh.elements.shape[0]
        # GET BOUNDARY EDGES
        reedges = tmesh.GetEdgesQuad()
        del tmesh

        tedges = time()-tedges
        #------------------------------------------------------------------------------------------

    class nmesh(object):
        # """Construct pMesh"""
        points = repoints
        elements = reelements
        edges = np.array([[],[]])
        faces = np.array([[],[]])
        nnode = repoints.shape[0]
        nelem = reelements.shape[0]
        info = 'hex'
    if ComputeAll is True:
        nmesh.edges = reedges
        nmesh.faces = refaces

    gc.collect()


    # print '\nHigh order meshing timing:\n\t\tElement loop:\t '+str(telements)+' seconds\n\t\tNode loop:\t\t '+str(tnodes)+\
    #  ' seconds'+'\n\t\tEdge loop:\t\t '+str(tedges)+' seconds'+\
    #  '\n\t\tFace loop:\t\t '+str(tfaces)+' seconds\n'

    return nmesh